import pytest
import pathlib

import numpy as np

from matsim import Vehicle

HERE = pathlib.Path(__file__).parent

files = ['output_allVehicles.xml.gz']


@pytest.mark.parametrize('filepath', files)
def test_vehicle_reader(filepath):
    vehicle_dataframes = Vehicle.vehicle_reader(HERE / filepath)
    
    vehicle_types = vehicle_dataframes.vehicle_types
    vehicles = vehicle_dataframes.vehicles
    vehicles_counts_dataframes = vehicles['type'].value_counts()
    
    EXPECTED_VEHICLE_TYPES_COLUMNS = ['id', 'accessTimeInSecondsPerPerson', 'doorOperationMode', 'egressTimeInSecondsPerPerson', 'seats', 'standingRoomInPersons', 'length', 'width', 'pce', 'networkMode', 'factor']
    EXPECTED_VEHICLES_COLUMNS = ['id','type']
    EXPECTED_VEHICLETYPES_ROW_0_RESULT = {
        'id': 'Bus',
        'accessTimeInSecondsPerPerson': 0.5,
        'doorOperationMode': 'serial',
        'egressTimeInSecondsPerPerson': 0.5,
        'seats': 70,
        'standingRoomInPersons': 0,
        'length': 18.0,
        'width': 2.5,
        'pce': 2.8,
        'networkMode': 'car',
        'factor': 1.0
    }
        
    # Checking total lengths
    assert len(vehicle_types) == 4
    assert len(vehicles) == 113
    
    # Checking vehicles types number of occurrences
    assert vehicles_counts_dataframes.defaultVehicleType == 44
    assert vehicles_counts_dataframes.Bus == 43
    assert vehicles_counts_dataframes.Tram == 25
    assert vehicles_counts_dataframes.Rail == 1
    
    assert vehicle_types.iloc[0].to_dict() == EXPECTED_VEHICLETYPES_ROW_0_RESULT
    np.testing.assert_array_equal(EXPECTED_VEHICLE_TYPES_COLUMNS, vehicle_types.keys())
    np.testing.assert_array_equal(EXPECTED_VEHICLES_COLUMNS, vehicles.keys())


import pandas as pd
from pandas.testing import assert_frame_equal
from unittest import TestCase

import matsim.Network


class TestNetworkHandler(TestCase):

    def test(self):
        network = matsim.Network.read_network('tests/test_network.xml.gz')
        nodes = network.nodes
        links = network.links

        # we know how many links and nodes the test network has
        self.assertEqual(15, len(nodes))
        self.assertEqual(23, len(links))

        # lets check for present of from and to node of link #1
        self.assertTrue(len(nodes.node_id[nodes.node_id=='1']) == 1)
        self.assertTrue(len(nodes.node_id[nodes.node_id=='2']) == 1)

        fromNode = nodes[nodes.node_id == '1']
        self.assertEqual(-20000, fromNode.x[0])
        self.assertEqual(0, fromNode.y[0])

        # test link #1 with all its attributes
        self.assertTrue(len(links.link_id[links.link_id=='1']) == 1)
        link = links[links.link_id == '1']

        self.assertEqual('1', link.from_node[0])
        self.assertEqual('2', link.to_node[0])
        self.assertEqual('1', link.link_id[0])
        self.assertEqual(36000, link.capacity[0])
        self.assertEqual(10000, link.length[0])
        self.assertEqual(27.78, link.freespeed[0])
        self.assertEqual(1, link.permlanes[0])
        self.assertEqual('car,bike', link.modes[0])

    def test_attrs(self):
        network = matsim.Network.read_network('tests/test_network_attrs.xml.gz')
        expected_node_attrs = [
            ['1', 'meta:name', 'node-1'],
            ['2', 'meta:name', 'node-2'],
            ['3', 'meta:name', 'node-3'],
            ['4', 'meta:name', 'node-4'],
        ]
        expected_link_attrs = [
            ['1', 'meta:name', 'link-1'],
            ['2', 'meta:name', 'link-2'],
            ['3', 'meta:name', 'link-3'],
        ]
        node_attrs = network.node_attrs
        link_attrs = network.link_attrs

        assert_frame_equal(node_attrs, pd.DataFrame(data=expected_node_attrs, columns=['node_id', 'name', 'value']))
        assert_frame_equal(link_attrs, pd.DataFrame(data=expected_link_attrs, columns=['link_id', 'name', 'value']))



# -*- coding: utf-8 -*-

def positive_constraint(mode, val):
    """ Ensures parameter are positive """
    return max(0, val)


def negative_constraint(mode, val):
    """ Ensure parameter are negative """
    return min(0, val)



#!/usr/bin/env python

import argparse
import json
import os
import time
from dataclasses import dataclass
from functools import reduce
from random import Random

import numpy as np
import jax.numpy as jnp
import optax
from jax import grad, random
from requests import post
from tqdm import trange

URL = "http://localhost:%d"

METADATA = "network-opt-freespeed", "Optimize parameters for free-speed model. Server must be running before starting."


def as_list(array):
    return [float(x) for x in array]


def req(port, priority, rbl, traffic_light):
    req = {
        "priority": as_list(priority),
        "right_before_left": as_list(rbl),
        "traffic_light": as_list(traffic_light),
    }

    results = [post(URL % p, json=req).json() for p in port]

    data = {}
    for k in results[0]["data"].keys():
        data[k] = reduce(lambda x, y: x + y, [d["data"][k] for d in results])

    res = {
        "rmse": np.mean([x["rmse"] for x in results]),
        "mae": np.mean([x["mae"] for x in results]),
        "data": data
    }

    return req, res


@dataclass
class Model:
    name: str
    module: object
    optimizer: optax.adam
    opt_state: object
    params: jnp.array
    loss: callable


def setup(parser: argparse.ArgumentParser):
    parser.add_argument("--steps", type=int, help="Number of training steps", default=1500)
    parser.add_argument("--resume", help="File with parameters to to resume", default=None)
    parser.add_argument("--port", type=int, nargs="+", help="Port to connect on", default=[9090])
    parser.add_argument("--ref-model", required=False, default=None,
                        help="Use an integrated model instead of importing", choices=["tree", "germany"])
    parser.add_argument("--learning-rate", type=float, help="Start learning rate", default=1e-4)
    parser.add_argument("--batch-size", type=int, help="Batch size", default=128)
    parser.add_argument("--output", help="Output folder for params", default="output-params")


def main(args):
    batch_size = args.batch_size
    batches = 5

    models = {}
    resume = {}

    if args.ref_model == "tree":
        from .ref_model import tree as p
        rbl = tl = p
    elif args.ref_model == "germany":
        from .ref_model.germany import speedRelative_priority as p
        from .ref_model.germany import speedRelative_right_before_left as rbl
        from .ref_model.germany import speedRelative_traffic_light as tl
    else:
        # Import model, must be present on path
        from gen_code import speedRelative_priority as p
        from gen_code import speedRelative_right_before_left as rbl
        from gen_code import speedRelative_traffic_light as tl

    if args.resume:
        print("Resuming from", args.resume)
        with open(args.resume) as f:
            resume = json.load(f)

    for (module, name) in ((p, "priority"), (tl, "traffic_light"), (rbl, "right_before_left")):
        schedule = optax.exponential_decay(
            init_value=args.learning_rate, decay_rate=0.9,
            # Every 5% steps, decay 0.9
            transition_steps=int(batches * args.steps / 20), transition_begin=int(0.35 * args.steps * batches),
            staircase=False
        )

        optimizer = optax.adam(schedule)
        params = jnp.array(resume[name] if name in resume else module.params)
        opt_state = optimizer.init(params)

        models[name] = Model(
            name, module, optimizer, opt_state, params, grad(module.batch_loss)
        )

    r = Random(42)

    out = os.path.join(args.output, time.strftime("%Y%m%d-%H%M"))
    os.makedirs(out, exist_ok=True)

    print("Writing to", out)

    with trange(args.steps) as outer:

        # A simple update loop.
        for i in outer:
            params, result = req(args.port, models["priority"].params,
                                 models["right_before_left"].params, models["traffic_light"].params)

            name = "it%04d_mae_%.3f_rmse_%.3f.json" % (i, result["mae"], result["rmse"])

            with open(os.path.join(out, name), "w") as f:
                json.dump(params, f)

            outer.set_postfix(mae=result["mae"], rmse=result["rmse"])
            data = result["data"]

            for k, m in models.items():

                xs = [d["x"] for d in data[k]]
                xs = jnp.array(xs)

                ys = [d["yTrue"] for d in data[k]]
                ys = jnp.array(ys)

                for j in range(batches):
                    key = random.PRNGKey(r.getrandbits(31))

                    idx = random.choice(key, jnp.arange(0, xs.shape[0]), shape=(batch_size,))

                    grads = m.loss(m.params, xs[idx], ys[idx])

                    updates, m.opt_state = m.optimizer.update(grads, m.opt_state)
                    m.params = optax.apply_updates(m.params, updates)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog=METADATA[0], description=METADATA[1])
    setup(parser)
    main(parser.parse_args())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import glob

from typing import Any, Tuple

import geopandas
import numpy as np
import pandas as pd


def calc_adjusted_mode_share(sim: pd.DataFrame, survey: pd.DataFrame,
                             count_var: str = "trips", dist_var: str = "dist_group") -> Tuple[Any, pd.DataFrame]:
    """ This function can be used if the given input trip data has a different distance distribution than the survey data.
        It will rescale the survey data to match simulated data, which allows to calculate an adjusted overall mode share.

        :param sim: data frame containing shares for distance group and modes
        :param survey: data frame containing shares from survey data
        :param count_var: name of column containing the number of trips or share in 'sim'
        :param dist_var: name of the column holding the distance group information
        :return: tuple of optimization result and adjusted mode share
    """
    from scipy.optimize import minimize

    sagg = sim.groupby(dist_var).sum()
    sagg['share'] = sagg[count_var] / np.sum(sagg[count_var])

    # Rescale the distance groups of the survey data so that it matches the distance group distribution of the simulation
    # The overall mode share after this adjustment will the resulting adjusted mode share
    def f(x, result=False):
        adj = survey.copy()

        for i, t in enumerate(x):
            adj.loc[adj[dist_var] == sagg.index[i], "share"] *= t

        adj.share = adj.share / np.sum(adj.share)

        agg = adj.groupby(dist_var).sum()

        # Minimized difference between adjusted and simulated distribution
        err = sum((sagg.share - agg.share)**2)

        if result:
            return adj

        return err

    # One variable for each distance group
    x0 = np.ones(len(sagg.index)) / len(sagg.index)

    # Sum should be less than one
    cons = [{'type': 'ineq', 'fun': lambda x:  1 - sum(x)}]
    bnds = tuple((0, 1) for x in x0)

    res = minimize(f, x0, method='SLSQP', bounds=bnds, constraints=cons)

    df = f(res.x, True)

    return res, df


def read_trips_and_persons(run, transform_persons=None, transform_trips=None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """ Read trips and persons from run directory """

    # Return input as output
    # This allows to re-use input for the calc functions
    if type(run) is tuple and len(run) == 2:
        return run

    trips = glob.glob(run.rstrip("/") + "/*.output_trips.csv.gz")[0]
    persons = glob.glob(run.rstrip("/") + "/*.output_persons.csv.gz")[0]

    df = pd.read_csv(trips, sep=";",  dtype={"person": "str"})
    dfp = pd.read_csv(persons, sep=";", dtype={"person": "str"})

    gdf = geopandas.GeoDataFrame(dfp, 
            geometry=geopandas.points_from_xy(dfp.first_act_x, dfp.first_act_y)
    )

    if transform_persons is not None:
        gdf = transform_persons(gdf)

    df = pd.merge(df, gdf, how="inner", left_on="person", right_on="person")

    nans = df.main_mode.isnull()

    # use longest distance mode if there is no main mode
    df.loc[nans, "main_mode"] = df.loc[nans, "longest_distance_mode"]

    if transform_trips is not None:
        df = transform_trips(df)

    return df, gdf

def read_leg_stats(run : str, transform_persons=None, transform_legs=None):
    """ Reads leg statistic from run directory """

    legs = glob.glob(run.rstrip("/") + "/*.output_trips.csv.gz")[0]
    persons = glob.glob(run.rstrip("/") + "/*.output_persons.csv.gz")[0]

    df = pd.read_csv(legs, sep=";")
    dfp = pd.read_csv(persons, sep=";", index_col=0)

    gdf = geopandas.GeoDataFrame(dfp, 
            geometry=geopandas.points_from_xy(dfp.first_act_x, dfp.first_act_y)
    )

    if transform_persons is not None:
        gdf = transform_persons(gdf)

    if transform_legs is not None:
        df = transform_legs(df)

    return df

def calc_mode_share(run, transform_persons=None, transform_trips=None):
    """ Calculates the mode share from output directory """    
    df, _ = read_trips_and_persons(run, transform_persons, transform_trips)

    return df.groupby("main_mode").count()["trip_number"] / len(df)

def calc_mode_stats(run, attrs=[], 
                    dist_bins = [0, 1000, 2000, 5000, 10000, 20000, np.inf],
                    dist_labels = ["0 - 1000", "1000 - 2000", "2000 - 5000", "5000 - 10000", "10000 - 20000", "20000+"],
                    transform_persons=None, transform_trips=None) -> pd.DataFrame:    
    """ Calculate detailed mode statistics """

    trips, _ = read_trips_and_persons(run, transform_persons, transform_trips)
    trips["dist_group"] = pd.cut(trips.traveled_distance, dist_bins, labels=dist_labels, right=False)

    def aggr(x):
        data = dict(n=len(x), mean_dist=np.average(x.traveled_distance))               
        return pd.Series(data=data)

    aggr = trips.groupby(attrs + ["dist_group", "main_mode"], observed=True).apply(aggr)
    
    aggr["n"].fillna(0, inplace=True)

    aggr["share"] = aggr.n / aggr.n.sum()
    aggr["share"].fillna(0, inplace=True)

    return aggr


def calc_population_stats(run, attrs=[], transform_persons=None, transform_trips=None) -> pd.DataFrame:
    """ Calculate population statistics """

    trips, persons = read_trips_and_persons(run, transform_persons, transform_trips)

    def mode_usage(mode):
        def f(x):
            return (x == mode).any()        
        return f
    
    n_trips = trips.groupby("person").agg(n_trips=("trip_number", "max"))
    
    persons = persons.join(n_trips, on="person")
    persons.n_trips.fillna(0, inplace=True)

    def summarize(x):
        
        total = len(x)
        total_mobile = (x.n_trips > 0).sum()
        
        mobile = total_mobile / total        
        
        args = {"%s_user" % k: ("main_mode", mode_usage(k)) for k in set(trips.main_mode)}
        
        p_trips = trips[trips.person.isin(x.person)]
        
        mode_user = p_trips.groupby(["person"]).agg(**args)    
        joined = x.join(mode_user, on="person", how="inner")

        data = {}
        for c in mode_user.columns:
            share = joined[c].sum() / total_mobile
            data[c] = share    
        
        data.update({"mobile": mobile, "n": len(x), "avg_trips": np.average(x.n_trips)})
        
        return pd.Series(data=data)

    if attrs:
        aggr = persons.groupby(attrs).apply(summarize)
    else:
        aggr = summarize(persons)
        aggr = aggr.to_frame(0).T

    aggr["population_share"] = aggr.n / aggr.n.sum()
    
    return aggr

if __name__ == "__main__":

    def prepare_persons(df):
        df["age_group"] = pd.cut(df.age, [0, 18, 66, np.inf], labels=["0 - 17", "18 - 65", "65+"], right=False)
        return df

    #df = calc_mode_stats("/Users/rakow/Volumes/math-cluster/matsim-kelheim/calibration-v3/runs/010", attrs=["age_group"], transform_persons=prepare_persons)
    #print(df)

    df = calc_population_stats("/Users/rakow/Volumes/math-cluster/matsim-kelheim/calibration-v3/runs/010", attrs=["age_group"], transform_persons=prepare_persons)
    print(df)


import pytest
import pathlib
import pandas as pd
import numpy as np
from collections import defaultdict

from matsim import Plans

HERE = pathlib.Path(__file__).parent

files = ['plans_full.xml.gz', 'plans_empty.xml.gz']


@pytest.mark.parametrize('filepath', files)
def test_plan_reader(filepath):
    count_people = defaultdict(int)
    count_plans = defaultdict(int)
    
    plans = Plans.plan_reader(HERE / filepath)

    for person,plan in plans:
        count_people[person.attrib['id']] += 1
        if plan: count_plans[person.attrib['id']] += 1

    if filepath == 'plans_empty.xml.gz':
        assert len(count_people) == 3
        assert len(count_plans) == 0
    else: 
        assert len(count_people) == 3
        assert len(count_plans) == 3
        assert count_plans['100024301'] == 2


@pytest.mark.parametrize('filepath', files)
def test_plan_reader_selected_only(filepath):
    count_people = defaultdict(int)
    count_plans = defaultdict(int)
    
    plans = Plans.plan_reader(HERE / filepath, selected_plans_only = True)

    for person,plan in plans:
        print(person)
        count_people[person.attrib['id']] += 1
        if plan: count_plans[person.attrib['id']] += 1

    if filepath == 'plans_empty.xml.gz':
        assert len(count_people) == 3
        assert len(count_plans) == 0
    else: 
        assert len(count_people) == 3
        assert len(count_plans) == 3
        assert count_plans['100024301'] == 1
        

def test_non_existent():
    with pytest.raises(IOError):
        plans = Plans.plan_reader("not existing.xml")
        for _ in plans:
            pass

@pytest.mark.parametrize('filepath', files)
def test_plan_reader_dataframe(filepath):
    selected_plans_cases = [True, False]
    
    for selected_plans_only in selected_plans_cases:
        plans_dataframes = Plans.plan_reader_dataframe(HERE / filepath, selected_plans_only)
        
        persons = plans_dataframes.persons
        plans = plans_dataframes.plans
        activities = plans_dataframes.activities
        legs = plans_dataframes.legs
        routes = plans_dataframes.routes
        
        persons_expected_columns_full = ['id', 'home-activity-zone', 'subpopulation']
        persons_expected_columns_empty = ['id', 'age', 'district', 'homeId', 'homeX', 'homeY', 'sex']
        plans_expected_columns = ['id', 'person_id', 'score', 'selected']
        activites_expected_columns = ['id', 'plan_id', 'type', 'link', 'x', 'y', 'end_time', 'zoneId', 'max_dur', 'cemdapStopDuration_s']
        legs_expected_columns = ['id', 'plan_id', 'mode', 'dep_time']
        routes_expected_columns = ['id', 'leg_id', 'value', 'type', 'start_link', 'end_link', 'trav_time', 'distance', 'vehicleRefId']
        
        if filepath == 'plans_empty.xml.gz':
            assert len(persons) == 3
            assert len(plans) == 0
            assert len(activities) == 0
            assert len(legs) == 0
            assert len(routes) == 0
            np.testing.assert_array_equal(persons_expected_columns_empty, persons.keys())
            np.testing.assert_array_equal([], plans.keys())
            np.testing.assert_array_equal([], activities.keys())
            np.testing.assert_array_equal([], legs.keys())
            np.testing.assert_array_equal([], routes.keys())

        else:
            assert len(persons) == 3
            
            if selected_plans_only:
                assert len(plans) == 3
            else:
                assert len(plans) == 4
            
            assert len(activities) == 39
            assert len(legs) == 35
            assert len(routes) == 35
            np.testing.assert_array_equal(persons_expected_columns_full, persons.keys())
            np.testing.assert_array_equal(plans_expected_columns, plans.keys())
            np.testing.assert_array_equal(activites_expected_columns, activities.keys())
            np.testing.assert_array_equal(legs_expected_columns, legs.keys())
            np.testing.assert_array_equal(routes_expected_columns, routes.keys())


#!/usr/bin/env python

import random
from argparse import ArgumentParser
from os import makedirs
from os.path import join

import optuna
import sklearn.ensemble
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import KFold

from .features import build_datasets
from .models import create_regressor, model_to_java, model_to_py

METADATA = "network-train-model", "Train models for network capacity and freeflow speed."


def setup(parser: ArgumentParser):
    parser.add_argument("--n-trials", type=int, help="Number of trials", default=250)
    parser.add_argument("--name", help="Name of the model", default="Model")
    parser.add_argument("--package", help="Package name", default="org.matsim.prepare.network")
    parser.add_argument("--output", help="Output folder", default="gen_code")
    parser.add_argument("--network-features", type=str, nargs="+", help="Path to file with edge features", required=True)
    parser.add_argument("--input-intersections", type=str, nargs="+", help="Path to file with intersection results",
                        required=True)
    parser.add_argument("--input-routes", type=str, nargs="+", help="Path to file with route results.", required=True)
    parser.add_argument("--model-type", help="Type of model (features to use)", choices=["default", "extended"],
                        default="default")


def main(args):
    dfs = build_datasets(args.network_features, args.input_intersections, args.input_routes, args.model_type)
    targets = dfs.keys()

    def get(idx, t):
        if idx is not None:
            df = dfs[t].iloc[idx]
        else:
            df = dfs[t]

        return df.drop(columns=["target"]), df.target.to_numpy()

    scaler = {}

    for t in targets:
        _scaler = sklearn.preprocessing.StandardScaler(with_mean=True)

        df = get(None, t)[0]

        norm = ["length", "speed", "num_lanes"]
        if args.model_type == "full":
            norm += ["num_foes", "junction_inc_lanes"]

        scaler[t] = sklearn.compose.ColumnTransformer([
            ("scale", _scaler, [df.columns.get_loc(x) for x in norm])  # column indices
        ],
            remainder="passthrough"
        )

        scaler[t].fit(df)

    print("Model targets", targets)

    # %%

    def best_model(ms, t):
        errors = []
        for m in ms:
            X, y = get(None, t)
            X = scaler[t].transform(X)

            pred = m.predict(X)
            err = mean_absolute_error(y, pred)

            errors.append((m, err))

        errors = sorted(errors, key=lambda m: m[1])

        return errors[0]

    fold = KFold(n_splits=6, shuffle=True)

    classifier = {
        'mean',
        'XGBRFRegressor',
        'XGBRegressor',
        'RandomForestRegressor',
        'ExtraTreesRegressor',
        'LGBMRegressor',
        'DecisionTreeRegressor',
        'PassiveAggressiveRegressor',
        # More
        #   'SVR',
        #   'KernelSVC',
        #   'QLatticeRegressor',
        #   'LinearSVR',
        #   'Ridge',
        #   'SGDRegressor',
        #   'LogisticRegression',
        #   'AdaGradRegressor',
        #   'CDRegressor',
        #   'FistaRegressor',
        #   'SDCARegressor',
        #   'Lasso',
        #   'ElasticNet'
    }

    def objective(classifier_name, target):
        global model

        def _fn(trial):
            global model

            r = random.Random(42)

            random_state = r.getrandbits(31)

            seq = iter(fold.split(dfs[target]))

            error = 0
            i = 0

            candidates = []

            for train, test in seq:
                model = create_regressor(trial, classifier_name, random_state)

                candidates.append(model)

                X, y = get(train, target)
                X = scaler[t].transform(X)

                model.fit(X, y)

                Xval, yval = get(test, target)
                Xval = scaler[t].transform(Xval)

                pred = model.predict(Xval)

                error += mean_absolute_error(yval, pred)

                i += 1

            best = best_model(candidates, t)[0]

            return error / i

        return _fn

    def callback(study, trial):
        global best
        global model
        if study.best_trial == trial:
            best = model

    models = {}
    for t in targets:

        print("Training", t)

        models[t] = {}

        for m in classifier:
            print("Running model", m)

            study = optuna.create_study(sampler=optuna.samplers.TPESampler(seed=42), direction='minimize')
            study.optimize(objective(m, t), n_trials=args.n_trials, callbacks=[callback], show_progress_bar=True)

            models[t][m] = best

    for t in targets:
        print("#### ", t)

        m = best_model(models[t].values(), t)

        print("Best model", m)

        makedirs(args.output, exist_ok=True)

        with open(join(args.output, "__init__.py"), "w") as f:
            f.write("")

        name = args.name + "_" + t

        with open(join(args.output, args.name + ".java"), "w") as f:
            f.write("""package %(package)s;

import org.matsim.application.prepare.network.opt.FeatureRegressor;
import org.matsim.application.prepare.network.opt.NetworkModel;

public class %(name)s implements NetworkModel {
	@Override
	public FeatureRegressor capacity(String junctionType) {
		return switch (junctionType) {
			case "traffic_light" -> %(name)s_capacity_traffic_light.INSTANCE;
			case "right_before_left" -> %(name)s_capacity_right_before_left.INSTANCE;
			case "priority" -> %(name)s_capacity_priority.INSTANCE;
			default -> throw new IllegalArgumentException("Unknown type: " + junctionType);
		};
	}

	@Override
	public FeatureRegressor speedFactor(String junctionType) {
		return switch (junctionType) {
			case "traffic_light" -> %(name)s_speedRelative_traffic_light.INSTANCE;
			case "right_before_left" -> %(name)s_speedRelative_right_before_left.INSTANCE;
			case "priority" -> %(name)s_speedRelative_priority.INSTANCE;
			default -> throw new IllegalArgumentException("Unknown type: " + junctionType);
		};
	}
}
""" % dict(package=args.package, name=args.name))

        with open(join(args.output, name + ".java"), "w") as f:
            code = model_to_java(name, args.package, m[0], scaler[t], get(None, t)[0])
            f.write(code)

        with open(join(args.output, t + ".py"), "w") as f:
            code = model_to_py(t, m[0], scaler[t], get(None, t)[0])
            f.write("# -*- coding: utf-8 -*-\n")
            f.write("\"\"\"%s\nError: %f\"\"\"\n" % m)
            f.write(code)


if __name__ == "__main__":
    parser = ArgumentParser(prog=METADATA[0], description=METADATA[1])
    setup(parser)
    main(parser.parse_args())



# -*- coding: utf-8 -*-

import sys
from typing import Sequence, Callable, Dict, Literal
from functools import reduce
from collections import defaultdict

import numpy as np
import optuna
import pandas as pd

from .base import CalibratorBase, CalibrationInput
from .analysis import read_trips_and_persons


def get_sbb_params(config, group, value, mode):
    """ Return parameter set for specific group """

    if "SBBBehaviorGroups" not in config:
        config["SBBBehaviorGroups"] = {}

    if "behaviorGroup" not in config["SBBBehaviorGroups"]:
        config["SBBBehaviorGroups"]["behaviorGroup"] = []

    groups = config["SBBBehaviorGroups"]["behaviorGroup"]

    for b_group in groups:

        if b_group["personAttribute"] == group:

            for v in b_group["personGroupAttributeValues"]:
                if v["attributeValues"] == value:

                    for m in v["absoluteModeCorrections"]:
                        if m == mode:
                            return m

                    m = {"mode": mode}
                    v["absoluteModeCorrections"].append(m)
                    return m

            m = {"mode": mode}
            b_group["personGroupAttributeValues"].append(
                {"attributeValues": value, "absoluteModeCorrections": [m]}
            )
            return m

    m = {"mode": mode}
    groups.append({"personAttribute": group,
                   "personGroupAttributeValues":
                       [{"attributeValues": value, "absoluteModeCorrections": [m]}]})
    return m


def parse_group(p):
    p = p.strip("[]")
    return {x.split("=")[0]: x.split("=")[1] for x in p.split(",")}


class ASCGroupCalibrator(CalibratorBase):
    """ Calibrates the alternative specific for specific subpopulations """

    def __init__(self,
                 modes: Sequence[str],
                 initial: CalibrationInput,
                 target: CalibrationInput,
                 fixed_mode: str = "walk",
                 lr: Callable[[int, str, float, optuna.Trial, optuna.Study], float] = None,
                 constraints: Dict[str, Callable[[str, float], float]] = None,
                 calib_base: Literal['always', 'alternating', 'never'] = "always",
                 multi_groups: Sequence[str] = None,
                 corr_correction: float = 1,
                 config_format: Literal['default', 'sbb'] = "default"):
        """Abstract constructors for all calibrations. Usually the same parameters should be made available subclasses.

        :param modes: list of all relevant modes
        :param initial: dict of initial asc values
        :param target: target shares as csv or dict
        :param fixed_mode: the fixed mode with asc=0
        :param lr: learning rate schedule, will be called with (trial number, mode, update, trial, study)
        :param constraints: constraints for each mode, must return asc and will be called with original asc
        :param calib_base: how to calibrate the ungrouped base ASC, can be always, alternating (every 2nd iteration) or never
        :param multi_groups: Groups that contain multiple values, that will be split by ","
        :param corr_correction: factor to reduce learning rates for possibly correlated groups, 0 disables this correction
        :param config_format: use SBBBehaviorGroups for the parameter config
        """
        super().__init__(modes, initial, target, lr, constraints)

        self.fixed_mode = fixed_mode
        self.multi_groups = set(multi_groups)
        self.corr_correction = corr_correction
        self.config_format = config_format

        if "mode" not in self.target.columns and "main_mode" not in self.target.columns:
            raise ValueError("Target must have 'mode' or 'main_mode' column")

        if calib_base not in ("always", "alternating", "never"):
            raise ValueError("calib_base must be one of 'always', 'alternating' or 'never'")

        if config_format not in ("default", "sbb"):
            raise ValueError("config_format must be one of 'default' or 'sbb'")

        self.target = self.target.rename(columns={"value": "share", "main_mode": "mode"}, errors="ignore")

        if "share" not in self.target.columns:
            raise ValueError("Target must have 'share' column")

        self.groups = [t for t in self.target.columns if t not in ('mode', 'share', 'asc')]

        if calib_base != "never":
            self.base = self.get_group(self.target, None)
            self.base = self.base.groupby("mode").agg(target=("share", "sum"))
            self.alternate_base = calib_base == "alternating"

            if len(self.base) == 0:
                raise ValueError("Target must contain base without any attributes")
        else:
            self.base = None
            self.alternate_base = False

    def get_group(self, df, groups: dict = None, use_multi=False):
        """ Get data of one group"""
        # return result for empty groups
        if groups is None:
            idx = reduce(lambda x, y: x & y, [pd.isna(df[g]) for g in self.groups])
            return df[idx]

        if use_multi:
            idx = []
            for g, v in groups.items():
                if g in self.multi_groups:
                    vs = v.split(",")
                    idx.append(df[g].isin(vs))
                else:
                    idx.append(df[g] == v)

            idx = reduce(lambda x, y: x & y, idx)
        else:
            idx = reduce(lambda x, y: x & y, [df[g] == v for g, v in groups.items()])

        return df[idx]

    @property
    def name(self):
        return "asc"

    @property
    def num_targets(self):
        return len(self.base)

    def init_study(self, study: optuna.Trial):
        study.set_user_attr("modes", self.modes)
        study.set_user_attr("groups", self.groups)
        study.set_user_attr("fixed_mode", self.fixed_mode)

        print("Calibrating groups:", self.groups)
        print("Running with base target:", self.base)

    def update_config(self, study: optuna.Study, trial: optuna.Trial, prefix: str, config: dict):

        # Base ascs for each mode
        base = defaultdict(lambda: 0)

        if self.base is not None:
            for mode in self.modes:
                m = self.get_mode_params(config, mode)
                m["constant"] = trial.suggest_float(prefix + mode, sys.float_info.min, sys.float_info.max)
                base[mode] = m["constant"]

        step = 1
        # If all groups were fully correlated, update step needs to be divided by number of groups
        if self.corr_correction > 0:
            step = 1 / (len(self.groups) * self.corr_correction)

        # Grouped ascs
        for g in self.groups:
            for v in set(self.target[g]):
                if pd.isna(v):
                    continue

                attr = (str(g) + "=" + str(v))
                for mode in self.modes:
                    # Update constants
                    if self.config_format == "sbb":
                        param = "[%s]-%s" % (attr, mode)
                        p = trial.suggest_float(prefix + param, sys.float_info.min, sys.float_info.max)

                        # don't write certain values
                        if p == 0 and mode == self.fixed_mode:
                            continue

                        m = get_sbb_params(config, g, v, mode)

                        delta = (p - base[mode]) * step
                        m["deltaConstant"] = delta
                        trial.set_user_attr("%s_delta" % param, delta)
                    else:
                        raise ValueError("Currently only ssb config format is supported")

    def sample_initial(self, param: str) -> float:
        attr, _, mode = param.rpartition("-")

        if mode in self.initial.index:
            return self.initial.loc[mode]

        # TODO: Load from column with same format as input

        return 0

    def update_step(self, param: str, last_trial: optuna.Trial, completed: Sequence[optuna.Trial]) -> float:

        if param == self.fixed_mode:
            return 0

        # Base mode shares
        if not param.startswith("["):
            # No update every 2nd iteration
            if self.alternate_base and len(completed) % 2 == 1:
                return 0

            return self.calc_asc_update(self.base.loc[param].target, last_trial.user_attrs["%s_share" % param],
                                        self.base.loc[self.fixed_mode].target,
                                        last_trial.user_attrs["%s_share" % self.fixed_mode])

        else:
            p, _, mode = param.rpartition("-")

            # No update every other iteration
            if self.alternate_base and len(completed) % 2 == 0:
                return 0

            if mode == self.fixed_mode:
                return 0

            t = self.get_group(self.target, parse_group(p)).set_index("mode")

            # TODO: the previous applied correction might be considered here as well

            return self.calc_asc_update(t.loc[mode].share, last_trial.user_attrs["%s_share" % param],
                                        t.loc[self.fixed_mode].share,
                                        last_trial.user_attrs["%s-%s_share" % (p, self.fixed_mode)])

    def calc_stats(self, trial: optuna.Trial, run_dir: str,
                   transform_persons: Callable = None,
                   transform_trips: Callable = None) -> Sequence[float]:

        trips, persons = read_trips_and_persons(run_dir,
                                                transform_persons=transform_persons,
                                                transform_trips=transform_trips)

        if self.base is not None:
            base_share = trips.groupby("main_mode").count()["trip_number"] / len(trips)
            base_share = self.base.merge(base_share, left_index=True, right_index=True).rename(
                columns={"trip_number": "share"})

            base_share["mae"] = np.abs(base_share.share - base_share.target)

            for kv in base_share.itertuples():
                trial.set_user_attr("%s_share" % kv.Index, kv.share)
                trial.set_user_attr("%s_mae" % kv.Index, kv.mae)

            print("Obtained base shares:")
            print(base_share)

        errs = []

        for g in self.groups:
            for v in set(self.target[g]):
                if pd.isna(v):
                    continue

                sub_trips = self.get_group(trips, {g: v}, g in self.multi_groups)
                sub_share = sub_trips.groupby("main_mode").count()["trip_number"] / len(sub_trips)
                sub_share.rename("share", inplace=True)

                if len(sub_trips) == 0:
                    print("Empty subgroup %s=%s" % (g, v), file=sys.stderr)

                target = self.get_group(self.target, {g: v}).rename(columns={"share": "target"})
                target = target.merge(sub_share, how="outer", left_on="mode", right_index=True)
                target.share.fillna(0, inplace=True)

                target["mae"] = np.abs(target.share - target.target)

                attr = (str(g) + "=" + str(v))
                for kv in target.itertuples():
                    trial.set_user_attr("[%s]-%s_share" % (attr, kv.mode), kv.share)
                    trial.set_user_attr("[%s]-%s_mae" % (attr, kv.mode), kv.mae)

                errs.append(target)

        errs = pd.concat(errs)
        print("Grouped sum of errors:")
        for g in self.groups:
            sub_err = errs.groupby(g).agg(sum_mae=("mae", "sum"))
            print(sub_err)

            for e in sub_err.itertuples():
                trial.set_user_attr("%s=%s_sum_mae" % (g, e.Index), e.sum_mae)

        return base_share.mae.to_numpy()



#!/usr/bin/env python

import os

from argparse import ArgumentParser

from .features import read_edges, read_intersections, read_routes

METADATA = "sumo-collect-results", "Aggregate results from sumo runs"


def setup(parser: ArgumentParser):

    parser.add_argument("mode", nargs='?', help="Convert result file that create with one of the run scripts",
                        choices=["edges", "intersections", "routes"])
    parser.add_argument("--input", help="Path to input file for conversion", required=True)

def main(args):

    df = None

    if args.mode == "edges":
        df = read_edges(args.input)

    elif args.mode == "intersections":
        df = read_intersections(args.input)

    elif args.mode == "routes":
        df = read_routes(args.input)

    if df is not None:
        base = os.path.basename(args.input.rstrip("/"))
        df.to_csv(f"result_{args.mode}_{base}.csv", index=False)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: matsim/pb/Events.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from matsim.pb import Ids_pb2 as matsim_dot_pb_dot_Ids__pb2
from matsim.pb import Model_pb2 as matsim_dot_pb_dot_Model__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16matsim/pb/Events.proto\x12\tmatsim.pb\x1a\x13matsim/pb/Ids.proto\x1a\x15matsim/pb/Model.proto\".\n\nEventBatch\x12 \n\x06\x65vents\x18\x01 \x03(\x0b\x32\x10.matsim.pb.Event\"\xaa\x07\n\x05\x45vent\x12\x0c\n\x04time\x18\x01 \x01(\x01\x12%\n\x06\x63oords\x18\x02 \x01(\x0b\x32\x15.matsim.pb.Coordinate\x12*\n\x07generic\x18\x03 \x01(\x0b\x32\x17.matsim.pb.GenericEventH\x00\x12\x32\n\x0b\x61\x63tivityEnd\x18\x04 \x01(\x0b\x32\x1b.matsim.pb.ActivityEndEventH\x00\x12\x36\n\ractivityStart\x18\x05 \x01(\x0b\x32\x1d.matsim.pb.ActivityStartEventH\x00\x12.\n\tlinkEnter\x18\x06 \x01(\x0b\x32\x19.matsim.pb.LinkEnterEventH\x00\x12.\n\tlinkLeave\x18\x07 \x01(\x0b\x32\x19.matsim.pb.LinkLeaveEventH\x00\x12\x38\n\x0fpersonalArrival\x18\x08 \x01(\x0b\x32\x1d.matsim.pb.PersonArrivalEventH\x00\x12:\n\x0fpersonDeparture\x18\t \x01(\x0b\x32\x1f.matsim.pb.PersonDepartureEventH\x00\x12\x42\n\x13personEntersVehicle\x18\n \x01(\x0b\x32#.matsim.pb.PersonEntersVehicleEventH\x00\x12\x42\n\x13personLeavesVehicle\x18\x0b \x01(\x0b\x32#.matsim.pb.PersonLeavesVehicleEventH\x00\x12\x32\n\x0bpersonMoney\x18\x0c \x01(\x0b\x32\x1b.matsim.pb.PersonMoneyEventH\x00\x12\x32\n\x0bpersonStuck\x18\r \x01(\x0b\x32\x1b.matsim.pb.PersonStuckEventH\x00\x12\x42\n\x13transitDriverStarts\x18\x0e \x01(\x0b\x32#.matsim.pb.TransitDriverStartsEventH\x00\x12\x36\n\rvehicleAborts\x18\x0f \x01(\x0b\x32\x1d.matsim.pb.VehicleAbortsEventH\x00\x12\x44\n\x14vehicleEntersTraffic\x18\x10 \x01(\x0b\x32$.matsim.pb.VehicleEntersTrafficEventH\x00\x12\x44\n\x14vehicleLeavesTraffic\x18\x11 \x01(\x0b\x32$.matsim.pb.VehicleLeavesTrafficEventH\x00\x42\x06\n\x04type\"}\n\x0cGenericEvent\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x31\n\x05\x61ttrs\x18\x02 \x03(\x0b\x32\".matsim.pb.GenericEvent.AttrsEntry\x1a,\n\nAttrsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x95\x01\n\x10\x41\x63tivityEndEvent\x12\"\n\x06linkId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12&\n\nfacilityId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12$\n\x08personId\x18\x03 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x0f\n\x07\x61\x63ttype\x18\x04 \x01(\t\"\x97\x01\n\x12\x41\x63tivityStartEvent\x12\"\n\x06linkId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12&\n\nfacilityId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12$\n\x08personId\x18\x03 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x0f\n\x07\x61\x63ttype\x18\x04 \x01(\t\"[\n\x0eLinkEnterEvent\x12\"\n\x06linkId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12%\n\tvehicleId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"[\n\x0eLinkLeaveEvent\x12\"\n\x06linkId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12%\n\tvehicleId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"o\n\x12PersonArrivalEvent\x12\"\n\x06linkId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x0f\n\x07legMode\x18\x02 \x01(\t\x12$\n\x08personId\x18\x03 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"q\n\x14PersonDepartureEvent\x12\"\n\x06linkId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x0f\n\x07legMode\x18\x02 \x01(\t\x12$\n\x08personId\x18\x03 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"g\n\x18PersonEntersVehicleEvent\x12$\n\x08personId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12%\n\tvehicleId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"g\n\x18PersonLeavesVehicleEvent\x12$\n\x08personId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12%\n\tvehicleId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"u\n\x10PersonMoneyEvent\x12$\n\x08personId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x01\x12\x0f\n\x07purpose\x18\x03 \x01(\t\x12\x1a\n\x12transactionPartner\x18\x04 \x01(\t\"m\n\x10PersonStuckEvent\x12$\n\x08personId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\"\n\x06linkId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x0f\n\x07legMode\x18\x03 \x01(\t\"\xe7\x01\n\x18TransitDriverStartsEvent\x12$\n\x08\x64riverId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12%\n\tvehicleId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12*\n\x0etransitRouteId\x18\x03 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12)\n\rtransitLineId\x18\x04 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\'\n\x0b\x64\x65partureId\x18\x05 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"_\n\x12VehicleAbortsEvent\x12%\n\tvehicleId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\"\n\x06linkId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\"\xc1\x01\n\x19VehicleEntersTrafficEvent\x12$\n\x08\x64riverId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\"\n\x06linkId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12%\n\tvehicleId\x18\x03 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x13\n\x0bnetworkMode\x18\x04 \x01(\t\x12\x1e\n\x16relativePositionOnLink\x18\x05 \x01(\x01\"\xc1\x01\n\x19VehicleLeavesTrafficEvent\x12$\n\x08\x64riverId\x18\x01 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\"\n\x06linkId\x18\x02 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12%\n\tvehicleId\x18\x03 \x01(\x0b\x32\x12.matsim.pb.ProtoId\x12\x13\n\x0bnetworkMode\x18\x04 \x01(\t\x12\x1e\n\x16relativePositionOnLink\x18\x05 \x01(\x01\x42\'\n\x18org.matsim.core.utils.pbB\x0bProtoEventsb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'matsim.pb.Events_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030org.matsim.core.utils.pbB\013ProtoEvents'
  _globals['_GENERICEVENT_ATTRSENTRY']._options = None
  _globals['_GENERICEVENT_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_EVENTBATCH']._serialized_start=81
  _globals['_EVENTBATCH']._serialized_end=127
  _globals['_EVENT']._serialized_start=130
  _globals['_EVENT']._serialized_end=1068
  _globals['_GENERICEVENT']._serialized_start=1070
  _globals['_GENERICEVENT']._serialized_end=1195
  _globals['_GENERICEVENT_ATTRSENTRY']._serialized_start=1151
  _globals['_GENERICEVENT_ATTRSENTRY']._serialized_end=1195
  _globals['_ACTIVITYENDEVENT']._serialized_start=1198
  _globals['_ACTIVITYENDEVENT']._serialized_end=1347
  _globals['_ACTIVITYSTARTEVENT']._serialized_start=1350
  _globals['_ACTIVITYSTARTEVENT']._serialized_end=1501
  _globals['_LINKENTEREVENT']._serialized_start=1503
  _globals['_LINKENTEREVENT']._serialized_end=1594
  _globals['_LINKLEAVEEVENT']._serialized_start=1596
  _globals['_LINKLEAVEEVENT']._serialized_end=1687
  _globals['_PERSONARRIVALEVENT']._serialized_start=1689
  _globals['_PERSONARRIVALEVENT']._serialized_end=1800
  _globals['_PERSONDEPARTUREEVENT']._serialized_start=1802
  _globals['_PERSONDEPARTUREEVENT']._serialized_end=1915
  _globals['_PERSONENTERSVEHICLEEVENT']._serialized_start=1917
  _globals['_PERSONENTERSVEHICLEEVENT']._serialized_end=2020
  _globals['_PERSONLEAVESVEHICLEEVENT']._serialized_start=2022
  _globals['_PERSONLEAVESVEHICLEEVENT']._serialized_end=2125
  _globals['_PERSONMONEYEVENT']._serialized_start=2127
  _globals['_PERSONMONEYEVENT']._serialized_end=2244
  _globals['_PERSONSTUCKEVENT']._serialized_start=2246
  _globals['_PERSONSTUCKEVENT']._serialized_end=2355
  _globals['_TRANSITDRIVERSTARTSEVENT']._serialized_start=2358
  _globals['_TRANSITDRIVERSTARTSEVENT']._serialized_end=2589
  _globals['_VEHICLEABORTSEVENT']._serialized_start=2591
  _globals['_VEHICLEABORTSEVENT']._serialized_end=2686
  _globals['_VEHICLEENTERSTRAFFICEVENT']._serialized_start=2689
  _globals['_VEHICLEENTERSTRAFFICEVENT']._serialized_end=2882
  _globals['_VEHICLELEAVESTRAFFICEVENT']._serialized_start=2885
  _globals['_VEHICLELEAVESTRAFFICEVENT']._serialized_end=3078
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Command line tool remove iterations directories. See help usage for details:
    >> python3.9 clean_iters.py -h
"""

import os
import re
import shutil
from argparse import ArgumentParser

METADATA = "clean-iters", "Remove all iterations directories except the last one"


def setup(parser: ArgumentParser):
    parser.add_argument("dirs", nargs="+", help="Directories to scan")
    parser.add_argument("-f", "--force", action='store_true', default=False, help="Don't ask for confirmation")
    parser.add_argument("-d", "--dry-run", action='store_true', default=False,
                        help="Don't remove anything, only print for information")


def main(args):
    pattern = re.compile(r"it\.(\d+)")

    for d in args.dirs:
        for dirpath, dirnames, filenames in os.walk(os.path.abspath(d)):

            matches = [pattern.match(it) for it in dirnames]

            iters = [(m.group(0), int(m.group(1))) for m in matches if m is not None]
            iters = sorted(iters, key=lambda k: k[1])

            if len(iters) > 1:

                print("Found %d iterations in %s" % (len(iters), dirpath))

                delete = False
                if args.dry_run:
                    delete = False
                elif args.force:
                    delete = True
                else:
                    answer = input("Delete [yn]?: ")
                    if any(answer.lower() == f for f in ["yes", 'y', '1', 'ye']):
                        delete = True
                    elif any(answer.lower() == f for f in ["no", 'n', '']):
                        delete = False
                    else:
                        print("Unknown answer, don't delete")

                if delete:

                    # Delete all except last
                    for iter, n in iters[:-1]:
                        path = os.path.join(dirpath, iter)
                        try:
                            shutil.rmtree(path)
                        except Exception as e:
                            print("Error removing %s:" % path, e)


if __name__ == "__main__":
    parser = ArgumentParser(prog=METADATA[0], description=METADATA[1])

    setup(parser)

    args = parser.parse_args()
    main(args)



# -*- coding: utf-8 -*-

import sys
from typing import Sequence, Callable, Dict

import optuna

from .base import CalibratorBase, CalibrationInput, to_float
from .analysis import calc_mode_share


class ASCCalibrator(CalibratorBase):
    """ Calibrates the alternative specific constant for desired modes """

    def __init__(self,
                 modes: Sequence[str],
                 initial: CalibrationInput,
                 target: CalibrationInput,
                 fixed_mode: str = "walk",
                 lr: Callable[[int, str, float, optuna.Trial, optuna.Study], float] = None,
                 constraints: Dict[str, Callable[[str, float], float]] = None):
        """Constructor

        :param modes: list of all relevant modes
        :param initial: dict of initial asc values
        :param target: target shares as csv or dict
        :param fixed_mode: the fixed mode with asc=0
        :param lr: learning rate schedule, will be called with (trial number, mode, update, trial, study)
        :param constraints: constraints for each mode, must return asc and will be called with original asc
        """
        super().__init__(modes, initial, target, lr, constraints)

        self.fixed_mode = fixed_mode

        self.target = self.target.rename(columns={"value": "share"}, errors="ignore")
        self.target = self.target.groupby("mode").agg(share=("share", "sum"))

    @property
    def name(self):
        return "asc"

    def init_study(self, study: optuna.Trial):
        study.set_user_attr("modes", self.modes)
        study.set_user_attr("fixed_mode", self.fixed_mode)

        print("Running study with target:", self.target)

    def update_config(self, study: optuna.Study, trial: optuna.Trial, prefix: str, config: dict):

        # Update constants
        for mode in self.modes:
            m = self.get_mode_params(config, mode)
            m["constant"] = trial.suggest_float(prefix + mode, sys.float_info.min, sys.float_info.max)

    def sample_initial(self, param: str) -> float:
        if param in self.initial.index:
            return self.initial.loc[param]

        return 0

    def update_step(self, param: str, last_trial: optuna.Trial, completed: Sequence[optuna.Trial]) -> float:

        if param == self.fixed_mode:
            return 0

        step = self.calc_asc_update(self.target.loc[param].share, last_trial.user_attrs["%s_share" % param],
                                    self.target.loc[self.fixed_mode].share,
                                    last_trial.user_attrs["%s_share" % self.fixed_mode])

        return step

    def calc_stats(self, trial: optuna.Trial, run_dir: str,
                   transform_persons: Callable = None,
                   transform_trips: Callable = None) -> Sequence[float]:

        shares = calc_mode_share(run_dir, transform_persons=transform_persons, transform_trips=transform_trips)
        print("Obtained mode shares:", shares)

        for k, v in shares.items():
            trial.set_user_attr("%s_share" % k, v)
            trial.set_user_attr("%s_mae" % k, to_float(abs(self.target.loc[k] - v)))

        return [to_float(abs(self.target.loc[mode] - trial.user_attrs["%s_share" % mode])) for mode in self.modes]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.utils import shuffle

from . import *


def prepare_persons(hh, pp, tt, augment=5, max_hh_size=5, core_weekday=False, remove_with_invalid_trips=False):
    """ Cleans common data errors and fill missing values """
    df = pp.join(hh, on="hh_id", lsuffix="hh_")

    # Augment data using p_weight
    if augment > 1:
        df = augment_persons(df, augment)
        df = shuffle(df, random_state=0).reset_index(drop=True)

    # set car avail
    df.loc[df.age < 17, "driving_license"] = Availability.NO
    fill(df, "driving_license", Availability.UNKNOWN)

    df["car_avail"] = (df.n_cars > 0) & (df.driving_license == Availability.YES)
    df["bike_avail"] = (df.n_bikes > 0) | (df.bike_avail == Availability.YES)

    # small children don't have pt abo
    df.loc[df.age < 6, "pt_abo_avail"] = Availability.NO
    fill(df, "pt_abo_avail", Availability.UNKNOWN)

    # Replace unknown income group
    fill(df, "economic_status", EconomicStatus.UNKNOWN)

    # Large households are underrepresented and capped
    df.n_persons = np.minimum(df.n_persons, max_hh_size)

    if core_weekday:
        df = df[df.reporting_day <= 4]

    if remove_with_invalid_trips:
        invalid = set(tt[~tt.valid].p_id)
        df = df[~df.p_id.isin(invalid)]

        mobile = set(tt[tt.valid].p_id)

        # Filter persons that are supposed to be mobile but have no trips
        df = df[(df.p_id.isin(mobile) | (~df.mobile_on_day))]

    df = df.drop(columns=['hh_id', 'present_on_day', 'reporting_day', 'location', 'h_weight',
                          'n_cars', 'n_bikes', 'n_other_vehicles', 'car_parking'])

    # Move the region type variable to the front because it is used as conditional
    df.insert(3, 'region_type', df.pop('region_type'))

    # Regions other than 1 and 3 are massively under-represented
    # Regions are reduced to 1 (Berlin) and 3 (Outside Berlin)
    df.loc[df.region_type != 1, "region_type"] = 3

    # Maximum age is 99, also to be able to encode age with two tokens
    df.loc[df.age >= 99, "age"] = 99

    return df

def bins_to_labels(bins):
    """ Convert bins to labels """
    res =  ["%.0f - %.0f" % (bins[i], bins[i + 1]) for i in range(len(bins) - 1)]

    if bins[-1] == np.inf:
        res[-1] = "%.0f+" % bins[-2]

    return res

def cut(x, bins):
    """ Cut x into bind and return labels """
    return pd.cut(x, bins, labels=bins_to_labels(bins), right=False)

def augment_persons(pp, factor=1, permute_age=0.5):
    """ Augment persons using p weight

    :param pp Person table
    :param factor multiply weight by this factor
    :param permute_age if larger 0 add gaussian noise to the age with parameter as scale
    """
    pp = pp.reset_index()

    duplicated = pp.loc[pp.index.repeat(np.maximum(1, np.rint(pp.p_weight * factor)))]

    if permute_age > 0:
        np.random.seed(0)

        age_permute = np.rint(np.random.normal(0, permute_age, len(duplicated))).astype(np.int32)
        duplicated.age += age_permute

        # 0 age is minimum
        duplicated.age = np.maximum(0, duplicated.age)

    # Filter invalid options
    return duplicated[check_age_employment(None, duplicated)]


def prepare_trips(pp, trips, core_weekday=True):
    """ Create trip data frame """

    df = trips.join(pp, on="p_id")

    # remove person if any trip is invalid
    invalid = set(trips[~trips.valid].p_id)
    df = df[~df.p_id.isin(invalid)]

    df = df.drop(columns=["t_weight", "sd_group", "mobile_on_day", "valid"])

    if core_weekday:
        # Monday - Thursday
        df = df[df.day_of_week <= 4]
        df = df.drop(columns=["day_of_week"])

    # reverse columns so activities are at the end
    return df[df.columns[::-1]]


def fill(df, col, val=None):
    """ Fill null values with dist of the rest (or replace val)"""
    if val is not None:
        df.loc[df[col] == val, col] = None

    isnull = df[col].isnull()
    sample = df[col].dropna().sample(isnull.sum(), replace=True).values
    df.loc[isnull, col] = sample


def create_activities(all_persons: pd.DataFrame, tt: pd.DataFrame, core_weekday=True, cut_groups=True,
                      include_person_context=True):
    """ Create activity representation from trip table """

    tt = tt.reset_index().set_index("p_id")

    import multiprocess as mp

    def convert(persons):
        res = []

        for p in persons.itertuples():

            acts = []
            valid = True

            # Filter if not done yet
            if "present_on_day" in persons.keys() and not p.present_on_day:
                continue

            p_id = p.p_id if "p_id" in persons.keys() else p.Index

            try:
                #            trips = tt[tt.p_id == p_id]
                trips = tt.loc[[p_id]]
            except KeyError:
                # Empty dataframe of same type
                trips = tt.iloc[:0, :].copy()

            if (~trips.valid).any():
                continue

            if core_weekday:
                # Monday - Thursday
                if (trips.day_of_week > 4).any():
                    continue

            # id generator
            def a_id(t_i):
                return "%s_%d" % (p.Index, t_i)

            if len(trips) == 0:
                acts.append(Activity(a_id(0), p.Index, 0, Purpose.HOME, 1440, 0, 0, TripMode.OTHER))
            else:
                acts.append(
                    Activity(a_id(0), p.Index, 0, trips.iloc[0].sd_group.source(), trips.iloc[0].departure, 0, 0,
                             TripMode.OTHER))

            for i in range(len(trips) - 1):
                t0 = trips.iloc[i]
                t1 = trips.iloc[i + 1]

                duration = t1.departure - t0.departure - t0.duration

                if duration < 0 or t0.gis_length < 0:
                    valid = False

                acts.append(Activity(a_id(i + 1), p.Index, i + 1, t0.purpose,
                                     duration, t0.gis_length, t0.duration, t0.main_mode))

            if len(trips) > 1:
                i += 1
                # last trip
                tl = trips.iloc[i]

                if tl.gis_length < 0:
                    valid = False

                # Duration is set to rest of day
                acts.append(
                    Activity(a_id(i + 1), p.Index, i + 1, tl.purpose, 1440, tl.gis_length, tl.duration, tl.main_mode))

            if valid:
                res.extend(acts)

        return res

    with mp.Pool(8) as pool:
        docs = pool.map(convert, np.array_split(all_persons, 16))
        result = functools.reduce(lambda a, b: a + b, docs)

    activities = pd.DataFrame(result).set_index("a_id")
    # Reverse columns because it will be reversed again at the end
    activities = activities.iloc[:, ::-1]
    persons = all_persons.iloc[:, ::-1].drop(columns=["p_id"], errors="ignore")

    if include_person_context:
        df = activities.join(persons, on="p_id", rsuffix="_p")
    else:
        df = activities

    df = df.drop(columns=["mobile_on_day", "p_weight", "hh_id", "present_on_day"], errors="ignore")

    # TODO: permute length and dist

    if cut_groups:
        df.duration = DurationGroup.cut(df.duration)
        df.leg_dist = DistanceGroup.cut(df.leg_dist)

    # reverse columns so activities are at the end
    return df.iloc[:, ::-1]


def check_age_employment(column_names, df):
    """ Check if age vs. employment status is valid"""

    invalid = (df.employment == Employment.CHILD) & (df.age > 7)
    invalid |= (df.employment == Employment.RETIREE) & (df.age < 30)
    invalid |= (df.employment == Employment.STUDENT) & ((df.age < 17) | (df.age > 55))
    invalid |= (df.employment == Employment.SCHOOL) & ((df.age < 5) | (df.age > 25))
    invalid |= (df.employment == Employment.UNEMPLOYED) & ((df.age < 17) | (df.age > 66))
    invalid |= (df.employment == Employment.JOB_PART_TIME) & (df.age < 18)
    invalid |= (df.employment == Employment.JOB_FULL_TIME) & (df.age < 18)
    invalid |= (df.employment == Employment.HOMEMAKER) & (df.age < 18)
    invalid |= (df.employment == Employment.TRAINEE) & (df.age < 16)

    return ~invalid


def check_age_license(column_names, df):
    """ Check driving license and age is valid """

    invalid = (df.driving_license == Availability.YES) & (df.age < 17)
    invalid |= df.car_avail & (df.age < 17)

    return ~invalid


def check_positive(column_names, df):
    """ Check for plausible number of vehicles"""

    invalid = (df.n_persons <= 0)
    #  invalid |= (df.n_cars < 0)
    #  invalid |= (df.n_bikes < 0)
    #  invalid |= (df.n_other_vehicles < 0)

    invalid |= (df.n_persons <= 1) & (df["type"] != HouseholdType.SINGLE)

    return ~invalid


def calc_commute(pp, tt):
    """ Calculate commute distances """

    work = tt[tt.valid &
              ((tt.sd_group == SourceDestinationGroup.HOME_WORK) | (tt.sd_group == SourceDestinationGroup.WORK_HOME))]

    edu = tt[tt.valid &
             ((tt.sd_group == SourceDestinationGroup.HOME_EDU) | (tt.sd_group == SourceDestinationGroup.EDU_HOME))]

    # t_weight is always the same for one person
    return work.groupby("p_id").agg(commute_dist=("gis_length", "mean"), weight=("t_weight", "max")), \
        edu.groupby("p_id").agg(commute_dist=("gis_length", "mean"), weight=("t_weight", "max"))


def calc_needed_short_distance_trips(ref_trips: pd.DataFrame, sim_trips: pd.DataFrame, max_dist=1000) -> Tuple[float, int]:
    """ Calculate number of short distance trips needed to add to match required share """

    target_share = float(ref_trips[ref_trips.gis_length < (max_dist / 1000)].t_weight.sum() / ref_trips.t_weight.sum())

    short_trips = sim_trips[sim_trips.traveled_distance < max_dist]

    current_share = len(short_trips) / len(sim_trips)
    num_trips = (len(short_trips) - len(sim_trips) * target_share) / (target_share - 1)

    return target_share, num_trips


import pathlib

from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()
VERSION = (HERE / "VERSION").read_text()

# This call to setup() does all the work
setup(
    version=VERSION,
    name="matsim-tools",
    description="MATSim Agent-Based Transportation Simulation Framework - official python tools",
    long_description_content_type="text/markdown",
    url="https://github.com/matsim-vsp/matsim-python-tools",
    author="VSP-Berlin",
    author_email="charlton@tu-berlin.de",
    license="GPLv3",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3",
    ],
    packages=["matsim"] + ["matsim." + x for x in find_packages(where="matsim")],
    install_requires=[
        "protobuf >= 3.20.0",
        "xopen >= 1.7.0",
        "pandas >= 2.1.0",
    ],
    extras_require={
        'calibration': ["optuna >= 3.3.0", "shapely >= 1.8.0", "geopandas >= 0.6.0"],
        # m2cgen has problems with newer xgb, see this issue
        # https://github.com/BayesWitnesses/m2cgen/issues/581
        'scenariogen': ["sumolib", "traci", "lxml", "optax", "requests", "tqdm", "scikit-learn", "xgboost==1.7.1", "lightgbm",
                        "sklearn-contrib-lightning", "numpy", "sympy", "m2cgen", "shapely", "optuna"]
    },
    tests_require=["assertpy", "pytest"],
    entry_points={
        'console_scripts': [
            'matsim-tools=matsim.cli.main:main',
            'matsim-scenariogen=matsim.scenariogen:main'
        ]
    },
    long_description=README,
)



from . import Events, Network, Plans, Vehicle, Facility, Household, TripEventHandler, writers

read_network = Network.read_network
event_reader = Events.event_reader
plan_reader = Plans.plan_reader
plan_reader_dataframe = Plans.plan_reader_dataframe
vehicle_reader = Vehicle.vehicle_reader
facility_reader = Facility.facility_reader
household_reader = Household.houshold_reader



#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.dev/sumo
# Copyright (C) 2010-2023 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    randomTrips.py
# @author  Daniel Krajzewicz
# @author  Jakob Erdmann
# @author  Michael Behrisch
# @date    2010-03-06


from __future__ import print_function
from __future__ import absolute_import
import os
import sys
import random
import bisect
import subprocess
from collections import defaultdict
import math

import sumolib  # noqa
from sumolib.miscutils import euclidean, parseTime, intIfPossible  # noqa
from sumolib.geomhelper import naviDegree, minAngleDegreeDiff  # noqa
from sumolib.net.lane import is_vehicle_class  # noqa

DUAROUTER = sumolib.checkBinary('duarouter')

SOURCE_SUFFIX = ".src.xml"
DEST_SUFFIX = ".dst.xml"
VIA_SUFFIX = ".via.xml"

MAXIMIZE_FACTOR = "max"


def get_options(args=None):
    op = sumolib.options.ArgumentParser(description="Generate trips between random locations",
                                        allowed_programs=['duarouter'])
    # input
    op.add_argument("-n", "--net-file", category="input", dest="netfile", required=True, type=op.net_file,
                    help="define the net file (mandatory)")
    op.add_argument("-a", "--additional-files", category="input", dest="additional", type=op.additional_file,
                    help="define additional files to be loaded by the router")
    op.add_argument("--weights-prefix", category="input", dest="weightsprefix", type=op.file,
                    help="loads probabilities for being source, destination and via-edge from the files named " +
                    "'prefix'.src.xml, 'prefix'.dst.xml and 'prefix'.via.xml")
    # output
    op.add_argument("-o", "--output-trip-file", category="output", dest="tripfile", type=op.route_file,
                    default="trips.trips.xml",
                    help="define the output trip filename")
    op.add_argument("-r", "--route-file", category="output", dest="routefile", type=op.route_file,
                    help="generates route file with duarouter")
    op.add_argument("--vtype-output", category="output", dest="vtypeout", type=op.file,
                    help="Store generated vehicle types in a separate file")
    op.add_argument("--weights-output-prefix", category="output", dest="weights_outprefix", type=op.file,
                    help="generates weights files for visualisation")
    # persons
    op.add_argument("--pedestrians", category="persons", action="store_true", default=False,
                    help="create a person file with pedestrian trips instead of vehicle trips")
    op.add_argument("--personrides", category="persons",
                    help="create a person file with rides using STR as lines attribute")
    op.add_argument("--persontrips", category="persons", action="store_true", default=False,
                    help="create a person file with person trips instead of vehicle trips")
    op.add_argument("--persontrip.transfer.car-walk", category="persons", dest="carWalkMode",
                    help="Where are mode changes from car to walking allowed " +
                    "(possible values: 'ptStops', 'allJunctions' and combinations)")
    op.add_argument("--persontrip.walkfactor", category="persons", dest="walkfactor", metavar="FLOAT", type=float,
                    help="Use FLOAT as a factor on pedestrian maximum speed during intermodal routing")
    op.add_argument("--persontrip.walk-opposite-factor", category="persons", dest="walkoppositefactor",
                    metavar="FLOAT", type=float,
                    help="Use FLOAT as a factor on pedestrian maximum speed against vehicle traffic direction")
    op.add_argument("--from-stops", category="persons", dest="fromStops",
                    help="Create trips that start at stopping places of the indicated type(s). i.e. 'busStop'")
    op.add_argument("--to-stops", category="persons", dest="toStops",
                    help="Create trips that end at stopping places of the indicated type(s). i.e. 'busStop'")
    # attributes
    op.add_argument("--prefix", category="attributes", dest="tripprefix", default="",
                    help="prefix for the trip ids")
    op.add_argument("-t", "--trip-attributes", category="attributes", dest="tripattrs", default="",
                    help="additional trip attributes. When generating pedestrians, attributes for " +
                    "'person' and 'walk' are supported.")
    op.add_argument("--fringe-start-attributes", category="attributes", dest="fringeattrs", default="",
                    help="additional trip attributes when starting on a fringe.")
    op.add_argument("--vehicle-class",
                    help="The vehicle class assigned to the generated trips (adds a standard vType definition " +
                    "to the output file).")
    op.add_argument("--random-departpos", category="attributes", dest="randomDepartPos", action="store_true",
                    default=False, help="Randomly choose a position on the starting edge of the trip")
    op.add_argument("--random-arrivalpos", category="attributes", dest="randomArrivalPos", action="store_true",
                    default=False, help="Randomly choose a position on the ending edge of the trip")
    op.add_argument("--junction-taz", category="attributes", dest="junctionTaz", action="store_true", default=False,
                    help="Write trips with fromJunction and toJunction")
    # weights
    op.add_argument("-l", "--length", category="weights", action="store_true", default=False,
                    help="weight edge probability by length")
    op.add_argument("-L", "--lanes", category="weights", action="store_true", default=False,
                    help="weight edge probability by number of lanes")
    op.add_argument("--edge-param", category="weights", dest="edgeParam",
                    help="use the given edge parameter as factor for edge")
    op.add_argument("--speed-exponent", category="weights", dest="speed_exponent", metavar="FLOAT", type=float,
                    default=0.0, help="weight edge probability by speed^'FLOAT' (default 0)")
    op.add_argument("--fringe-speed-exponent", category="weights", dest="fringe_speed_exponent", metavar="FLOAT",
                    help="weight fringe edge probability by speed^'FLOAT' (default: speed exponent)")
    op.add_argument("--angle", category="weights", dest="angle", default=90.0, type=float,
                    help="weight edge probability by angle [0-360] relative to the network center")
    op.add_argument("--angle-factor", category="weights", dest="angle_weight", default=1.0, type=float,
                    help="maximum weight factor for angle")
    op.add_argument("--random-factor", category="weights", dest="randomFactor", default=1.0, type=float,
                    help="edge weights are dynamically disturbed by a random factor drawn uniformly from [1,FLOAT]")
    op.add_argument("--fringe-factor", category="weights", dest="fringe_factor", default="1.0",
                    help="multiply weight of fringe edges by 'FLOAT' (default 1)" +
                    " or set value 'max' to force all traffic to start/end at the fringe.")
    op.add_argument("--fringe-threshold", category="weights", dest="fringe_threshold", default=0.0, type=float,
                    help="only consider edges with speed above 'FLOAT' as fringe edges (default 0)")
    op.add_argument("--allow-fringe", category="weights", dest="allow_fringe", action="store_true", default=False,
                    help="Allow departing on edges that leave the network and arriving on edges " +
                    "that enter the network (via turnarounds or as 1-edge trips")
    op.add_argument("--allow-fringe.min-length", category="weights", dest="allow_fringe_min_length", type=float,
                    help="Allow departing on edges that leave the network and arriving on edges " +
                    "that enter the network, if they have at least the given length")
    op.add_argument("--fringe-junctions", category="weights", action="store_true", dest="fringeJunctions",
                    default=False, help="Determine fringe edges based on junction attribute 'fringe'")
    op.add_argument("--vclass", "--edge-permission", category="weights", default="passenger",
                    help="only from and to edges which permit the given vehicle class")
    op.add_argument("--via-edge-types", category="weights", dest="viaEdgeTypes",
                    help="Set list of edge types that cannot be used for departure or arrival " +
                    "(unless being on the fringe)")
    op.add_argument("--allow-roundabouts", category="weights", dest="allowRoundabouts", action="store_true",
                    default=False, help="Permit trips that start or end inside a roundabout")
    # processing
    op.add_argument("-s", "--seed", default=42, type=int,
                    help="random seed")
    op.add_argument("--random", action="store_true", default=False,
                    help="use a random seed to initialize the random number generator")
    op.add_argument("--min-distance", dest="min_distance", metavar="FLOAT", default=0.0,
                    type=float, help="require start and end edges for each trip to be at least 'FLOAT' m apart")
    op.add_argument("--min-distance.fringe", dest="min_dist_fringe", metavar="FLOAT", type=float,
                    help="require start and end edges for each fringe to fringe trip to be at least 'FLOAT' m apart")
    op.add_argument("--max-distance", dest="max_distance", metavar="FLOAT", type=float,
                    help="require start and end edges for each trip to be at most 'FLOAT' m " +
                    "apart (default 0 which disables any checks)")
    op.add_argument("-i", "--intermediate", default=0, type=int,
                    help="generates the given number of intermediate way points")
    op.add_argument("--jtrrouter", action="store_true", default=False,
                    help="Create flows without destination as input for jtrrouter")
    op.add_argument("--maxtries", default=100, type=int,
                    help="number of attemps for finding a trip which meets the distance constraints")
    op.add_argument("--remove-loops", dest="remove_loops", action="store_true", default=False,
                    help="Remove loops at route start and end")
    op.add_argument("--random-routing-factor", dest="randomRoutingFactor", default=1, type=float,
                    help="Edge weights for routing are dynamically disturbed "
                    "by a random factor drawn uniformly from [1,FLOAT)")
    op.add_argument("--validate", default=False, action="store_true",
                    help="Whether to produce trip output that is already checked for connectivity")
    op.add_argument("-v", "--verbose", action="store_true", default=False,
                    help="tell me what you are doing")
    # flow
    op.add_argument("-b", "--begin", category="flow", default=0, type=op.time,
                    help="begin time")
    op.add_argument("-e", "--end", category="flow", default=3600, type=op.time,
                    help="end time (default 3600)")
    group = op.add_mutually_exclusive_group()
    group.add_argument("-p", "--period", nargs="+", metavar="FLOAT", category="flow",
                       action=sumolib.options.SplitAction,
                       help="Generate vehicles with equidistant departure times and period=FLOAT (default 1.0). " +
                       "If option --binomial is used, the expected arrival rate is set to 1/period.")
    group.add_argument("--insertion-rate", dest="insertionRate", nargs="+", metavar="FLOAT", category="flow",
                       action=sumolib.options.SplitAction,
                       help="How much vehicles arrive in the simulation per hour (alternative to the period option).")
    group.add_argument("--insertion-density", dest="insertionDensity", nargs="+", metavar="FLOAT", category="flow",
                       action=sumolib.options.SplitAction,
                       help="How much vehicles arrive in the simulation per hour per kilometer of road " +
                       "(alternative to the period option).")
    op.add_argument("--flows", category="flow", default=0, type=int,
                    help="generates INT flows that together output vehicles with the specified period")
    op.add_argument("--random-depart", category="flow", action="store_true", dest="randomDepart", default=False,
                    help="Distribute departures randomly between begin and end")
    op.add_argument("--binomial",  category="flow", metavar="N", type=int,
                    help="If this is set, the number of departures per second will be drawn from a binomial " +
                    "distribution with n=N and p=PERIOD/N where PERIOD is the argument given to --period")

    try:
        options = op.parse_args(args=args)
    except (NotImplementedError, ValueError) as e:
        print(e, file=sys.stderr)
        sys.exit(1)

    if options.vclass:
        if not is_vehicle_class(options.vclass):
            print("Error: The string '%s' doesn't correspond to a legit vehicle class" % options.vclass,
                  file=sys.stderr)
            sys.exit(1)

    if options.persontrips or options.personrides:
        options.pedestrians = True

    if options.pedestrians:
        options.vclass = 'pedestrian'
        if options.flows > 0:
            print("Error: Person flows are not supported yet", file=sys.stderr)
            sys.exit(1)

    if options.validate and options.routefile is None:
        options.routefile = "routes.rou.xml"

    if options.period is None and options.insertionRate is None and options.insertionDensity is None:
        options.period = [1.]

    options.net = sumolib.net.readNet(options.netfile)
    if options.insertionDensity:
        # Compute length of the network
        length = 0.  # In meters
        for edge in options.net.getEdges():
            if edge.allows(options.vclass):
                length += edge.getLaneNumber() * edge.getLength()
        options.insertionRate = [density * (length / 1000.0) for density in options.insertionDensity]

    if options.insertionRate:
        options.period = [3600.0 / rate for rate in options.insertionRate]

    if options.period:
        if any(options.period) <= 0:
            print("Error: Period must be positive", file=sys.stderr)
            sys.exit(1)
        options.period = list(map(intIfPossible, options.period))
        if options.binomial is not None:
            for p in options.period:
                if 1.0 / p / options.binomial >= 1:
                    print(("Warning: Option --binomial %s is too low for insertion period %s." % (
                           options.binomial, p))
                          + " Insertions will not be randomized.", file=sys.stderr)

    if options.jtrrouter and options.flows <= 0:
        print("Error: Option --jtrrouter must be used with option --flows", file=sys.stderr)
        sys.exit(1)

    if options.vehicle_class:
        if not is_vehicle_class(options.vehicle_class):
            print("Error: The string '%s' doesn't correspond to a legit vehicle class" %
                  options.vehicle_class, file=sys.stderr)
            sys.exit(1)

        if options.tripprefix:
            options.vtypeID = "%s_%s" % (options.tripprefix, options.vehicle_class)
        else:
            options.vtypeID = options.vehicle_class

        if 'type=' in options.tripattrs:
            print("Error: trip-attribute 'type' cannot be used together with option --vehicle-class", file=sys.stderr)
            sys.exit(1)

    if options.randomDepartPos:
        if 'departPos' in options.tripattrs:
            print("Error: trip-attribute 'departPos' cannot be used together with option --random-departpos",
                  file=sys.stderr)
            sys.exit(1)

    if options.randomArrivalPos:
        if 'arrivalPos' in options.tripattrs:
            print("Error: trip-attribute 'arrivalPos' cannot be used together with option --random-arrivalpos",
                  file=sys.stderr)
            sys.exit(1)

    if options.weightsprefix:
        weight_files = [options.weightsprefix + s for s in (SOURCE_SUFFIX, DEST_SUFFIX, VIA_SUFFIX)]
        if not any([os.path.isfile(w) for w in weight_files]):
            print("Error: None of the weight files '%s' exists." % "', '".join(weight_files), file=sys.stderr)
            sys.exit(1)

    if options.randomFactor < 1:
        print("Error: Option --random-factor requires a value >= 1.")
        sys.exit(1)

    if options.fromStops or options.toStops:
        options.edgeFromStops, options.edgeToStops = loadStops(options)

    if options.viaEdgeTypes:
        options.viaEdgeTypes = options.viaEdgeTypes.split(',')
    if options.fringe_speed_exponent is None:
        options.fringe_speed_exponent = options.speed_exponent

    if options.fringe_factor.lower() == MAXIMIZE_FACTOR:
        options.fringe_factor = MAXIMIZE_FACTOR
    else:
        try:
            options.fringe_factor = float(options.fringe_factor)
            if options.fringe_factor < 0:
                print("Error: --fringe-factor argument may not be negative", file=sys.stderr)
                sys.exit(1)
        except ValueError:
            print("Error: --fringe-factor argument must be a float or 'max'", file=sys.stderr)
            sys.exit(1)

    return options


class InvalidGenerator(Exception):
    pass


def loadStops(options):
    edgeFromStops = defaultdict(list)  # edge -> [(stopType1, stopID1), ...]
    edgeToStops = defaultdict(list)  # edge -> [(stopType1, stopID1), ...]
    if options.additional is None:
        print("Error: Option %s requires option --additional-files for loading infrastructure elements" %
              ("--from-stops" if options.fromStops else "--to-stops"), file=sys.stderr)
        sys.exit(1)
    stopTypes = []
    if options.fromStops:
        options.fromStops = options.fromStops.split(',')
        stopTypes += options.fromStops
    else:
        options.fromStops = []
    if options.toStops:
        options.toStops = options.toStops.split(',')
        stopTypes += options.toStops
    else:
        options.toStops = []
    stopTypes = list(set(stopTypes))
    typeCounts = defaultdict(lambda: 0)
    for additional in options.additional.split(','):
        for stop in sumolib.xml.parse(additional, stopTypes):
            edgeID = stop.lane.rsplit('_', 1)[0]
            if stop.name in options.fromStops:
                edgeFromStops[edgeID].append((stop.name, stop.id))
            if stop.name in options.toStops:
                edgeToStops[edgeID].append((stop.name, stop.id))
            typeCounts[stop.name] += 1

    if options.fromStops:
        available = sum([typeCounts[t] for t in options.fromStops])
        if available == 0:
            print("No stops of type%s '%s' were found in additional-files %s" % (
                ('' if len(options.fromStops) == 1 else 's'),
                options.fromStops[0], options.additional), file=sys.stderr)
            sys.exit(1)
    if options.toStops:
        available = sum([typeCounts[t] for t in options.toStops])
        if available == 0:
            print("No stops of type%s '%s' were found in additional-files %s" % (
                ('' if len(options.toStops) == 1 else 's'),
                options.toStops[0], options.additional), file=sys.stderr)
            sys.exit(1)
    return edgeFromStops, edgeToStops


# assigns a weight to each edge using weight_fun and then draws from a discrete
# distribution with these weights


class RandomEdgeGenerator:

    def __init__(self, net, weight_fun):
        self.net = net
        self.weight_fun = weight_fun
        self.cumulative_weights = []
        self.total_weight = 0
        for edge in self.net._edges:
            # print edge.getID(), weight_fun(edge)
            self.total_weight += weight_fun(edge)
            self.cumulative_weights.append(self.total_weight)
        if self.total_weight == 0:
            raise InvalidGenerator()

    def get(self):
        r = random.random() * self.total_weight
        index = bisect.bisect(self.cumulative_weights, r)
        return self.net._edges[index]

    def write_weights(self, fname, interval_id, begin, end):
        # normalize to [0,100]
        normalizer = 100.0 / max(1, max(map(self.weight_fun, self.net._edges)))
        weights = [(self.weight_fun(e) * normalizer, e.getID()) for e in self.net.getEdges()]
        weights.sort(reverse=True)
        with open(fname, 'w+') as f:
            f.write('<edgedata>\n')
            f.write('    <interval id="%s" begin="%s" end="%s">\n' % (
                interval_id, begin, end))
            for weight, edgeID in weights:
                f.write('        <edge id="%s" value="%0.2f"/>\n' %
                        (edgeID, weight))
            f.write('    </interval>\n')
            f.write('</edgedata>\n')


class RandomTripGenerator:

    def __init__(self, source_generator, sink_generator, via_generator, intermediate, pedestrians):
        self.source_generator = source_generator
        self.sink_generator = sink_generator
        self.via_generator = via_generator
        self.intermediate = intermediate
        self.pedestrians = pedestrians

    def get_trip(self, min_distance, max_distance, maxtries=100, junctionTaz=False, min_dist_fringe=None):
        for min_dist in [min_distance, min_dist_fringe]:
            if min_dist is None:
                break
            for _ in range(maxtries):
                source_edge = self.source_generator.get()
                intermediate = [self.via_generator.get() for __ in range(self.intermediate)]
                sink_edge = self.sink_generator.get()
                is_fringe2fringe = source_edge.is_fringe() and sink_edge.is_fringe() and not intermediate
                if min_dist == min_dist_fringe and not is_fringe2fringe:
                    continue
                if self.pedestrians:
                    destCoord = sink_edge.getFromNode().getCoord()
                else:
                    destCoord = sink_edge.getToNode().getCoord()
                coords = ([source_edge.getFromNode().getCoord()] +
                          [e.getFromNode().getCoord() for e in intermediate] +
                          [destCoord])
                distance = sum([euclidean(p, q)
                                for p, q in zip(coords[:-1], coords[1:])])
                if (distance >= min_dist
                        and (not junctionTaz or source_edge.getFromNode() != sink_edge.getToNode())
                        and (max_distance is None or distance < max_distance)):
                    return source_edge, sink_edge, intermediate
        raise Exception("Warning: no trip found after %s tries" % maxtries)


def get_prob_fun(options, fringe_bonus, fringe_forbidden, max_length):
    # fringe_bonus None generates intermediate way points
    randomProbs = defaultdict(lambda: 1)
    if options.randomFactor != 1:
        for edge in options.net.getEdges():
            randomProbs[edge.getID()] = random.uniform(1, options.randomFactor)

    roundabouts = set()
    if not options.allowRoundabouts:
        for roundabout in options.net.getRoundabouts():
            roundabouts.update(roundabout.getEdges())

    stopDict = None
    if options.fromStops and fringe_bonus == "_incoming":
        stopDict = options.edgeFromStops
    elif options.toStops and fringe_bonus == "_outgoing":
        stopDict = options.edgeToStops

    # IDEA: source and sink are not allowed as well, but that might also remove worthwile routes
    def edge_probability(edge):
        bonus_connections = None if fringe_bonus is None else getattr(edge, fringe_bonus)
        forbidden_connections = None if fringe_forbidden is None else getattr(edge, fringe_forbidden)
        if options.vclass and not edge.allows(options.vclass) and not stopDict:
            return 0  # not allowed
        if fringe_bonus is None and edge.is_fringe() and not options.pedestrians:
            return 0  # not suitable as intermediate way point
        if (fringe_forbidden is not None and
                edge.is_fringe(forbidden_connections) and
                not options.pedestrians and
                (options.allow_fringe_min_length is None or edge.getLength() < options.allow_fringe_min_length)):
            return 0  # the wrong kind of fringe
        if (fringe_bonus is not None and options.viaEdgeTypes is not None and
                not edge.is_fringe(bonus_connections, checkJunctions=options.fringeJunctions) and
                edge.getType() in options.viaEdgeTypes):
            return 0  # the wrong type of edge (only allows depart and arrival on the fringe)
        if fringe_bonus is not None and edge.getID() in roundabouts:
            return 0  # traffic typically does not start/end inside a roundabout
        prob = randomProbs[edge.getID()]
        if stopDict:
            prob *= len(stopDict[edge.getID()])
        if options.length:
            if (options.fringe_factor != 1.0 and fringe_bonus is not None and
                    edge.is_fringe(bonus_connections, checkJunctions=options.fringeJunctions)):
                # short fringe edges should not suffer a penalty
                prob *= max_length
            else:
                prob *= edge.getLength()
        if options.lanes:
            prob *= edge.getLaneNumber()
        if edge.is_fringe(bonus_connections, checkJunctions=options.fringeJunctions):
            prob *= (edge.getSpeed() ** options.fringe_speed_exponent)
        else:
            prob *= (edge.getSpeed() ** options.speed_exponent)
        if options.fringe_factor != 1.0 and fringe_bonus is not None:
            isFringe = (edge.getSpeed() > options.fringe_threshold and
                        edge.is_fringe(bonus_connections, checkJunctions=options.fringeJunctions))
            if isFringe and options.fringe_factor != MAXIMIZE_FACTOR:
                prob *= options.fringe_factor
            elif not isFringe and options.fringe_factor == MAXIMIZE_FACTOR:
                prob = 0
        if options.edgeParam is not None:
            prob *= float(edge.getParam(options.edgeParam, 1.0))
        if options.angle_weight != 1.0 and fringe_bonus is not None:
            xmin, ymin, xmax, ymax = edge.getBoundingBox()
            ex, ey = ((xmin + xmax) / 2, (ymin + ymax) / 2)
            nx, ny = options.angle_center
            edgeAngle = naviDegree(math.atan2(ey - ny, ex - nx))
            angleDiff = minAngleDegreeDiff(options.angle, edgeAngle)
            # print("e=%s nc=%s ec=%s ea=%s a=%s ad=%s" % (
            #    edge.getID(), options.angle_center, (ex,ey), edgeAngle,
            #    options.angle, angleDiff))
            # relDist = 2 * euclidean((ex, ey), options.angle_center) / max(xmax - xmin, ymax - ymin)
            # prob *= (relDist * (options.angle_weight - 1) + 1)
            if fringe_bonus == "_incoming":
                # source edge
                prob *= (angleDiff * (options.angle_weight - 1) + 1)
            else:
                prob *= ((180 - angleDiff) * (options.angle_weight - 1) + 1)

        return prob
    return edge_probability


class LoadedProps:

    def __init__(self, fname):
        self.weights = defaultdict(lambda: 0)
        for edge in sumolib.output.parse_fast(fname, 'edge', ['id', 'value']):
            self.weights[edge.id] = float(edge.value)

    def __call__(self, edge):
        return self.weights[edge.getID()]


def buildTripGenerator(net, options):
    try:
        max_length = 0
        for edge in net.getEdges():
            if not edge.is_fringe():
                max_length = max(max_length, edge.getLength())
        forbidden_source_fringe = None if options.allow_fringe else "_outgoing"
        forbidden_sink_fringe = None if options.allow_fringe else "_incoming"
        source_generator = RandomEdgeGenerator(
            net, get_prob_fun(options, "_incoming", forbidden_source_fringe, max_length))
        sink_generator = RandomEdgeGenerator(
            net, get_prob_fun(options, "_outgoing", forbidden_sink_fringe, max_length))
        if options.weightsprefix:
            if os.path.isfile(options.weightsprefix + SOURCE_SUFFIX):
                source_generator = RandomEdgeGenerator(
                    net, LoadedProps(options.weightsprefix + SOURCE_SUFFIX))
            if os.path.isfile(options.weightsprefix + DEST_SUFFIX):
                sink_generator = RandomEdgeGenerator(
                    net, LoadedProps(options.weightsprefix + DEST_SUFFIX))
    except InvalidGenerator:
        print("Error: no valid edges for generating source or destination. Try using option --allow-fringe",
              file=sys.stderr)
        return None

    try:
        via_generator = RandomEdgeGenerator(
            net, get_prob_fun(options, None, None, 1))
        if options.weightsprefix and os.path.isfile(options.weightsprefix + VIA_SUFFIX):
            via_generator = RandomEdgeGenerator(
                net, LoadedProps(options.weightsprefix + VIA_SUFFIX))
    except InvalidGenerator:
        if options.intermediate > 0:
            print("Error: no valid edges for generating intermediate points", file=sys.stderr)
            return None
        else:
            via_generator = None

    return RandomTripGenerator(
        source_generator, sink_generator, via_generator, options.intermediate, options.pedestrians)


def is_walk_attribute(attr):
    for cand in ['arrivalPos', 'speed=', 'duration=', 'busStop=']:
        if cand in attr:
            return True
    return False


def is_persontrip_attribute(attr):
    for cand in ['vTypes', 'modes']:
        if cand in attr:
            return True
    return False


def is_person_attribute(attr):
    for cand in ['departPos', 'type']:
        if cand in attr:
            return True
    return False


def is_vehicle_attribute(attr):
    # speedFactor could be used in vType and vehicle but we need it in the vType
    # to allow for the multi-parameter version
    for cand in ['depart', 'arrival', 'line', 'personNumber', 'containerNumber', 'type']:
        if cand in attr:
            return True
    return False


def split_trip_attributes(tripattrs, pedestrians, hasType, verbose):
    # handle attribute values with a space
    # assume that no attribute value includes an '=' sign
    allattrs = []
    for a in tripattrs.split():
        if "=" in a:
            allattrs.append(a)
        else:
            if len(allattrs) == 0:
                print("Warning: invalid trip-attribute '%s'" % a)
            else:
                allattrs[-1] += ' ' + a

    # figure out which of the tripattrs belong to the <person> or <vehicle>,
    # which belong to the <vType> and which belong to the <walk> or <persontrip>
    vehicleattrs = []
    personattrs = []
    vtypeattrs = []
    otherattrs = []
    for a in allattrs:
        if pedestrians:
            if is_walk_attribute(a) or is_persontrip_attribute(a):
                otherattrs.append(a)
            elif is_person_attribute(a):
                personattrs.append(a)
            else:
                vtypeattrs.append(a)
        else:
            if is_vehicle_attribute(a):
                vehicleattrs.append(a)
            else:
                vtypeattrs.append(a)

    if not hasType:
        if pedestrians:
            personattrs += vtypeattrs
        else:
            vehicleattrs += vtypeattrs
        vtypeattrs = []

    return (prependSpace(' '.join(vtypeattrs)),
            prependSpace(' '.join(vehicleattrs)),
            prependSpace(' '.join(personattrs)),
            prependSpace(' '.join(otherattrs)))


def prependSpace(s):
    if len(s) == 0 or s[0] == " ":
        return s
    else:
        return " " + s


def samplePosition(edge):
    return random.uniform(0.0, edge.getLength())


def main(options):
    if not options.random:
        random.seed(options.seed)

    if options.min_distance > options.net.getBBoxDiameter() * (options.intermediate + 1):
        options.intermediate = int(math.ceil(options.min_distance / options.net.getBBoxDiameter())) - 1
        print(("Warning: Using %s intermediate waypoints to achieve a minimum trip length of %s in a network "
               "with diameter %.2f.") % (options.intermediate, options.min_distance, options.net.getBBoxDiameter()),
              file=sys.stderr)

    if options.angle_weight != 1:
        xmin, ymin, xmax, ymax = options.net.getBoundary()
        options.angle_center = (xmin + xmax) / 2, (ymin + ymax) / 2

    trip_generator = buildTripGenerator(options.net, options)
    idx = 0

    vtypeattrs, options.tripattrs, personattrs, otherattrs = split_trip_attributes(
        options.tripattrs, options.pedestrians, options.vehicle_class, options.verbose)

    vias = {}

    time_delta = (parseTime(options.end) - parseTime(options.begin)) / len(options.period)
    times = [parseTime(options.begin) + i*time_delta for i in range(len(options.period)+1)]
    times = list(map(intIfPossible, times))

    def generate_origin_destination(trip_generator, options):
        source_edge, sink_edge, intermediate = trip_generator.get_trip(
            options.min_distance, options.max_distance, options.maxtries,
            options.junctionTaz, options.min_dist_fringe)
        return source_edge, sink_edge, intermediate

    def generate_attributes(idx, departureTime, arrivalTime, origin, destination, intermediate, options):
        label = "%s%s" % (options.tripprefix, idx)
        combined_attrs = options.tripattrs
        if options.randomDepartPos:
            randomPosition = samplePosition(origin)
            combined_attrs += ' departPos="%.2f"' % randomPosition
        if options.randomArrivalPos:
            randomPosition = samplePosition(destination)
            combined_attrs += ' arrivalPos="%.2f"' % randomPosition
        if options.fringeattrs and origin.is_fringe(
                origin._incoming, checkJunctions=options.fringeJunctions):
            combined_attrs += " " + options.fringeattrs
        if options.junctionTaz:
            attrFrom = ' fromJunction="%s"' % origin.getFromNode().getID()
            attrTo = ' toJunction="%s"' % destination.getToNode().getID()
        else:
            attrFrom = ' from="%s"' % origin.getID()
            attrTo = ' to="%s"' % destination.getID()
        if options.fromStops:
            attrFrom = ' %s="%s"' % random.choice(options.edgeFromStops[origin.getID()])
        if options.toStops:
            attrTo = ' %s="%s"' % random.choice(options.edgeToStops[destination.getID()])
        via = ""
        if intermediate:
            via = ' via="%s" ' % ' '.join(
                [e.getID() for e in intermediate])
            if options.validate:
                vias[label] = via
        return label, combined_attrs, attrFrom, attrTo, via

    def generate_one_person(label, combined_attrs, attrFrom, attrTo, departureTime, intermediate, options):
        fouttrips.write(
            '    <person id="%s" depart="%.2f"%s>\n' % (label, departureTime, personattrs))
        element = "walk"
        attrs = otherattrs
        if options.fromStops:
            fouttrips.write('        <stop%s duration="0"/>\n' % attrFrom)
            attrFrom = ''
        if options.persontrips:
            element = "personTrip"
        elif options.personrides:
            element = "ride"
            attrs = ' lines="%s%s"' % (options.personrides, otherattrs)
        if intermediate:
            fouttrips.write('        <%s%s to="%s"%s/>\n' % (element, attrFrom, intermediate[0].getID(), attrs))
            for edge in intermediate[1:]:
                fouttrips.write('        <%s to="%s"%s/>\n' % (element, edge.getID(), attrs))
            fouttrips.write('        <%s%s%s/>\n' % (element, attrTo, attrs))
        else:
            fouttrips.write('        <%s%s%s%s/>\n' % (element, attrFrom, attrTo, attrs))
        fouttrips.write('    </person>\n')

    def generate_one_flow(label, combined_attrs, departureTime, arrivalTime, period, options, timeIdx):
        if len(options.period) > 1:
            label = label + "#%s" % timeIdx
        if options.binomial:
            for j in range(options.binomial):
                fouttrips.write(('    <flow id="%s#%s" begin="%s" end="%s" probability="%.2f"%s/>\n') % (
                    label, j, departureTime, arrivalTime, 1.0 / period / options.binomial,
                    combined_attrs))
        else:
            fouttrips.write(('    <flow id="%s" begin="%s" end="%s" period="%s"%s/>\n') % (
                label, departureTime, arrivalTime, intIfPossible(period * options.flows), combined_attrs))

    def generate_one_trip(label, combined_attrs, departureTime):
        fouttrips.write('    <trip id="%s" depart="%.2f"%s/>\n' % (
            label, departureTime, combined_attrs))

    def generate_one(idx, departureTime, arrivalTime, period, origin, destination, intermediate, timeIdx=None):
        try:
            label, combined_attrs, attrFrom, attrTo, via = generate_attributes(
                idx, departureTime, arrivalTime, origin, destination, intermediate, options)

            if options.pedestrians:
                generate_one_person(label, combined_attrs, attrFrom, attrTo, departureTime, intermediate, options)
            else:
                if options.jtrrouter:
                    attrTo = ''

                combined_attrs = attrFrom + attrTo + via + combined_attrs

                if options.flows > 0:
                    generate_one_flow(label, combined_attrs, departureTime, arrivalTime, period, options, timeIdx)
                else:
                    generate_one_trip(label, combined_attrs, departureTime)

        except Exception as exc:
            if options.verbose:
                print(exc, file=sys.stderr)

        return idx + 1

    with open(options.tripfile, 'w') as fouttrips:
        sumolib.writeXMLHeader(fouttrips, "$Id$", "routes", options=options)
        if options.vehicle_class:
            vTypeDef = '    <vType id="%s" vClass="%s"%s/>\n' % (
                options.vtypeID, options.vehicle_class, vtypeattrs)
            if options.vtypeout:
                # ensure that trip output does not contain types, file may be
                # overwritten by later call to duarouter
                if options.additional is None:
                    options.additional = options.vtypeout
                else:
                    options.additional = options.additional + "," + options.vtypeout
                with open(options.vtypeout, 'w') as fouttype:
                    sumolib.writeXMLHeader(fouttype, "$Id$", "additional", options=options)
                    fouttype.write(vTypeDef)
                    fouttype.write("</additional>\n")
            else:
                fouttrips.write(vTypeDef)
            options.tripattrs += ' type="%s"' % options.vtypeID
            personattrs += ' type="%s"' % options.vtypeID

        if trip_generator:
            if options.flows == 0:
                for i in range(len(times)-1):
                    time = departureTime = parseTime(times[i])
                    arrivalTime = parseTime(times[i+1])
                    period = options.period[i]
                    if options.binomial is None:
                        departures = []
                        if options.randomDepart:
                            subsecond = math.fmod(period, 1)
                            while time < arrivalTime:
                                rTime = random.randrange(int(departureTime), int(arrivalTime))
                                time += period
                                if subsecond != 0:
                                    # allow all multiples of subsecond to appear
                                    rSubSecond = math.fmod(
                                        subsecond * random.randrange(int(departureTime), int(arrivalTime)), 1)
                                    rTime = min(arrivalTime, rTime + rSubSecond)
                                departures.append(rTime)
                            departures.sort()
                        else:
                            while departureTime < arrivalTime:
                                departures.append(departureTime)
                                departureTime += period

                        for time in departures:
                            # generate with constant spacing
                            try:
                                origin, destination, intermediate = generate_origin_destination(trip_generator, options)
                                idx = generate_one(idx, time, arrivalTime, period, origin, destination, intermediate)
                            except Exception as exc:
                                print(exc, file=sys.stderr)
                    else:
                        time = departureTime
                        while time < arrivalTime:
                            # draw n times from a Bernoulli distribution
                            # for an average arrival rate of 1 / period
                            prob = 1.0 / period / options.binomial
                            for _ in range(options.binomial):
                                if random.random() < prob:
                                    try:
                                        origin, destination, intermediate = generate_origin_destination(
                                            trip_generator, options)
                                        idx = generate_one(idx, time, arrivalTime, period,
                                                           origin, destination, intermediate)
                                    except Exception as exc:
                                        if options.verbose:
                                            print(exc, file=sys.stderr)
                            time += 1.0
            else:
                try:
                    origins_destinations = [generate_origin_destination(
                        trip_generator, options) for _ in range(options.flows)]
                    for i in range(len(times)-1):
                        for j in range(options.flows):
                            departureTime = times[i]
                            arrivalTime = times[i+1]
                            period = options.period[i]
                            origin, destination, intermediate = origins_destinations[j]
                            generate_one(j, departureTime, arrivalTime, period, origin, destination, intermediate, i)
                except Exception as exc:
                    if options.verbose:
                        print(exc, file=sys.stderr)

        fouttrips.write("</routes>\n")

    # call duarouter for routes or validated trips
    args = [DUAROUTER, '-n', options.netfile, '-r', options.tripfile, '--ignore-errors',
            '--begin', str(options.begin), '--end', str(options.end),
            '--alternatives-output', 'NUL',
            '--no-step-log']
    if options.additional is not None:
        args += ['--additional-files', options.additional]
    if options.carWalkMode is not None:
        args += ['--persontrip.transfer.car-walk', options.carWalkMode]
    if options.walkfactor is not None:
        args += ['--persontrip.walkfactor', str(options.walkfactor)]
    if options.walkoppositefactor is not None:
        args += ['--persontrip.walk-opposite-factor', str(options.walkoppositefactor)]
    if options.remove_loops:
        args += ['--remove-loops']
    if options.randomRoutingFactor != 1:
        args += ['--weights.random-factor', str(options.randomRoutingFactor)]
    if options.vtypeout is not None:
        args += ['--vtype-output', options.vtypeout]
    if options.junctionTaz:
        args += ['--junction-taz']
    if not options.verbose:
        args += ['--no-warnings']
    else:
        args += ['-v']

    options_to_forward = sumolib.options.get_prefixed_options(options)
    if 'duarouter' in options_to_forward:
        for option in options_to_forward['duarouter']:
            option[0] = '--' + option[0]
            if option[0] not in args:
                args += option
            else:
                raise ValueError("The argument '%s' has already been passed without the duarouter prefix." % option[0])

    if options.routefile:
        args2 = args + ['-o', options.routefile]
        if options.verbose:
            print("calling", " ".join(args2))
            sys.stdout.flush()
        subprocess.call(args2)
        sys.stdout.flush()
        sumolib.xml.insertOptionsHeader(options.routefile, options)

    if options.validate:
        # write to temporary file because the input is read incrementally
        tmpTrips = options.tripfile + ".tmp"
        args2 = args + ['-o', tmpTrips, '--write-trips']
        if options.junctionTaz:
            args2 += ['--write-trips.junctions']
        if options.verbose:
            print("calling", " ".join(args2))
            sys.stdout.flush()
        subprocess.call(args2)
        sys.stdout.flush()
        os.remove(options.tripfile)  # on windows, rename does not overwrite
        os.rename(tmpTrips, options.tripfile)
        sumolib.xml.insertOptionsHeader(options.tripfile, options)

    if trip_generator and options.weights_outprefix:
        idPrefix = ""
        if options.tripprefix:
            idPrefix = options.tripprefix + "."
        trip_generator.source_generator.write_weights(
            options.weights_outprefix + SOURCE_SUFFIX,
            idPrefix + "src", options.begin, options.end)
        trip_generator.sink_generator.write_weights(
            options.weights_outprefix + DEST_SUFFIX,
            idPrefix + "dst", options.begin, options.end)
        if trip_generator.via_generator:
            trip_generator.via_generator.write_weights(
                options.weights_outprefix + VIA_SUFFIX,
                idPrefix + "via", options.begin, options.end)

    # return wether trips could be generated as requested
    return trip_generator is not None


if __name__ == "__main__":
    if not main(get_options()):
        print("Error: Trips couldn't be generated as requested. "
              "Try the --verbose option to output more details on the failure.")
        sys.exit(1)



import xopen
import xml.etree.ElementTree as ET
import pandas as pd


class Plans:
    def __init__(self, persons, plans, activities, legs, routes):
        self.persons = persons
        self.plans = plans
        self.activities = activities
        self.legs = legs
        self.routes = routes

def plan_reader(filename, selected_plans_only = False):
    person = None
    tree = ET.iterparse(xopen.xopen(filename), events=['start','end'])
    
    for xml_event, elem in tree:
        if elem.tag == 'person' and xml_event == 'start':
            # keep track of whether a person node has any plans
            this_person_has_plans = False

            if person: person.clear() # clear memory
            person = elem

        elif elem.tag == 'plan' and xml_event == 'end':
            this_person_has_plans = True

            # filter out unselected plans if asked to do so
            if selected_plans_only and elem.attrib['selected'] == 'no': continue

            yield (person, elem)

            # free memory. Otherwise the data is kept in memory
            elem.clear()
        
        elif elem.tag == 'person' and xml_event == 'end':
            # if this person has no plans, then yield the person with a None plan.
            if not this_person_has_plans:
                yield (person, None)

# Parses attributes of an element and adds them to the given dictionary
def _parseAttributes(elem, dict):
    for attrib in elem.attrib:
        dict[attrib] = elem.attrib[attrib]
    return dict

# Returns dataframes with the following relations between them:
# Person : None
# Plan : person_id
# Activity : plan_id
# Leg : plan_id
# Route :leg_id
# The column names of the dataframes are the same as the attribute names (<name:'value'> and <attribute> are parsed)
def plan_reader_dataframe(filename, selected_plans_only = False):
    tree = ET.iterparse(xopen.xopen(filename), events=['start','end'])
    
    persons = []
    plans = []
    activities = []
    legs = []
    routes = []
    
    current_person = {}
    current_plan = {}
    current_activity = {}
    current_leg = {}
    current_route = {}
    
    # Indicates current parent element while parsing <attribute> element
    is_parsing_person = False
    is_parsing_activity = False
    is_parsing_leg = False
    
    current_person_id = None
    current_plan_id = 0
    current_activity_id = 0
    current_leg_id = 0
    current_route_id = 0
    
    for xml_event, elem in tree:
        if elem.tag in ['person', 'leg', 'activity', 'plan', 'route'] and xml_event == 'end':
            if is_parsing_person:
                persons.append(current_person)
                current_person = {}
                is_parsing_person = False
            
            if is_parsing_activity:
                activities.append(current_activity)
                current_activity = {}
                is_parsing_activity = False
            
            if is_parsing_leg:
                legs.append(current_leg)
                current_leg = {}
                is_parsing_leg = False
            
            if elem.tag == 'plan':
                if elem.attrib['selected'] == 'no' and selected_plans_only: continue
                plans.append(current_plan)
                current_plan = {}
                
            if elem.tag == 'route':
                routes.append(current_route)
                current_route = {}
            
            elem.clear()
        
        # PERSON
        elif elem.tag == 'person':
            current_person['id'] = elem.attrib['id']
            current_person_id = elem.attrib['id']
            is_parsing_person = True
        
        # PLAN
        elif elem.tag == 'plan':
            if elem.attrib['selected'] == 'no' and selected_plans_only: continue
            current_plan_id += 1
            
            current_plan['id'] = current_plan_id
            current_plan['person_id'] = current_person_id
            current_plan = _parseAttributes(elem, current_plan)
        
        # ACTIVITY
        elif elem.tag == 'activity':
            is_parsing_activity = True
            current_activity_id += 1
            
            current_activity['id'] = current_activity_id
            current_activity['plan_id'] = current_plan_id
            current_activity = _parseAttributes(elem, current_activity)
            
        
        # LEG
        elif elem.tag == 'leg':
            is_parsing_leg = True
            current_leg_id += 1
            
            current_leg['id'] = current_leg_id
            current_leg['plan_id'] = current_plan_id
            current_leg = _parseAttributes(elem, current_leg)
        
        
        # ROUTE
        elif elem.tag == 'route':
            current_route_id += 1
            
            current_route['id'] = current_route_id
            current_route['leg_id'] = current_leg_id
            current_route['value'] = elem.text
            current_route = _parseAttributes(elem, current_route)
        
        
        # ATTRIBUTES
        elif elem.tag == 'attribute' and xml_event == 'end':
            attribs = elem.attrib
            
            if is_parsing_activity:
                current_activity[attribs['name']] = elem.text
                
            elif is_parsing_leg:
                current_leg[attribs['name']] = elem.text
            
            elif is_parsing_person: # Parsing person
                current_person[attribs['name']] = elem.text
    
    persons = pd.DataFrame.from_records(persons)
    plans = pd.DataFrame.from_records(plans)
    activities = pd.DataFrame.from_records(activities)
    legs = pd.DataFrame.from_records(legs)
    routes = pd.DataFrame.from_records(routes)
    
    return Plans(persons, plans, activities, legs, routes)
    


# -*- coding: utf-8 -*-

import re

import pandas as pd
import seaborn as sns

from .utils import study_as_df


def plot_study(study):
    """ Create convergence plots for a study """

    modes = study.user_attrs["modes"]
    groups = study.user_attrs.get("groups", None)

    df = study_as_df(study)

    df_all = df[["trial"] + [x + "_mae" for x in modes]]

    df_all = pd.melt(df_all, "trial", var_name="main_mode", value_name="mae")
    df_all.main_mode = df_all.main_mode.str.removesuffix("_mae")

    ax = sns.lineplot(x="trial", y="mae", hue="main_mode", hue_order=modes, data=df_all)
    fig = ax.get_figure()

    fig.savefig("mae_all.png")

    if groups:

        for g in groups:

            p = re.compile(r"\[%s=(.+)\]-(.+)_mae" % g)

            columns = df.columns[df.columns.str.startswith("[" + g) & df.columns.str.endswith("_mae")]
            vars = [p.search(x) for x in columns]

            df_g = df[["trial"] + list(columns)]

            data = []

            # convert wide to long
            for idx, r in df_g.iterrows():
                for c, v in zip(columns, vars):
                    data.append({
                        "trial": r["trial"],
                        "mae": r[c],
                        g: v[1],
                        "main_mode": v[2]
                    })

            df_g = pd.DataFrame(data)

            grid = sns.FacetGrid(df_g, col=g)
            grid.map_dataframe(sns.lineplot, x="trial", y="mae", hue="main_mode", hue_order=modes)

            grid.figure.savefig("mae_%s.png" % g)



import xopen
import xml.etree.ElementTree as ET
import pandas as pd
from matsim import utils

class Facility:
    def __init__(self, facilities):
        self.facilities = facilities

# Returns facilities as a dataframe
def facility_reader(filename, convert_dataframes_types=True):
    tree = ET.iterparse(xopen.xopen(filename, 'r'), events=['start','end'])
    
    facilities = []
    current_facility = {}
    
    for xml_event, elem in tree:
        # FACILITY
        if elem.tag == 'facility':
            if xml_event == 'start':
                utils.parse_attributes(elem, current_facility)
            else:
                facilities.append(current_facility)
                current_facility = {}
                elem.clear()

        # EVERYTHING ELSE
        else:
            utils.parse_attributes(elem, current_facility)
            
    # Convert to dataframe and converts columns types
    facilities = pd.DataFrame.from_records(facilities)
    if convert_dataframes_types:
        try:
            facilities['x'] = facilities['x'].astype(float)
            facilities['y'] = facilities['y'].astype(float)
        except KeyError:
            print('dataframe types conversion failed')
    
    return Facility(facilities)